/***************************************************************************
                          actionmanager.cpp  -  eboxy action manager
                             -------------------
    begin                : Sun Dec 22 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <vector>
#include <iostream>

#include "actionmanager.h"
#include "../errors.h"

using namespace std;

// Allowed user actions
const char *useractions[] = {
  "add",
  "edit",
  "delete",
  "play",
  "record",
  "stop",
  "pause",
  "rewind",
  "fastforward",
  "jumptostart",
  "jumptoend",
  "slowspeed",
  "fastspeed",
  "eject",
  "nextchannel",
  "prevchannel",
  "frequencyup",
  "frequencydown",
  "volumeup",
  "volumedown",
  "mute",
  "shutdown",
  "restart",
  "restartserver",
  "up",
  "down",
  "left",
  "right",
  "zoomin",
  "zoomout",
  "yes",
  "no",
  "cancel",
  "tv",
  "radio",
  "rectimer",
  "dvd",
  "mail",
  "web",
  "photos",
  "videos",
  "music",
  "number0",
  "number1",
  "number2",
  "number3",
  "number4",
  "number5",
  "number6",
  "number7",
  "number8",
  "number9",
  "menu1",
  "menu2",
  "menu3",
  "viewmode",
  "nextaudio",
  "prevaudio",
  "nextsubtitle",
  "prevsubtitle",
  "nextangle",
  "prevangle",
  "deinterlace",
  "snapshot",
  "help",
  "trackinfo",
  "tracksearch",
  "systeminfo",
  "setup",
  "showclock",
  NULL  // end of list
};

ActionManager::ActionManager(void) {

}

ActionManager::~ActionManager(void) {
  // Release all handlers
  actionhandler *handler;
  while(localhandlers.size()>0) {
    handler = localhandlers.back();
    localhandlers.pop_back();
    delete handler;
  }
  while(globalhandlers.size()>0) {
    handler = globalhandlers.back();
    globalhandlers.pop_back();
    delete handler;
  }
}


int ActionManager::RegisterLocalAction(string actionname, ActionObject *aobject) {
  // Get a pointer to the actual action name string
  string nameref = GetUserAction(actionname);

  if(nameref.length()) {
    // Make sure the action isn't already registered
    for(unsigned int i=0; i<localhandlers.size(); i++) {
      if(localhandlers[i]->actionname == actionname) {
        cerr << "action '" << actionname << "' is already registered" << endl;
        return EBERR_ACTION_ASSIGNED;
      }
    }
    // Register the action
    actionhandler *handler = new actionhandler;
    handler->actionname = nameref;
    handler->aobject = aobject;
    localhandlers.push_back(handler);
    return EBERR_SUCCESS;
  }
  else {
    // Not a valid user action
    cerr << "'" << actionname << "' is not a valid user action" << endl;
    return EBERR_ACTION_INVALID;
  }
}

int ActionManager::RegisterGlobalAction(string actionname, ActionObject *aobject) {
  // Get a pointer to the actual action name string
  string nameref = GetUserAction(actionname);

  if(nameref.length()) {
    // Make sure the action isn't already registered
    for(unsigned int i=0; i<globalhandlers.size(); i++) {
      if(globalhandlers[i]->actionname == actionname) {
        cerr << "action '" << actionname << "' is already registered" << endl;
        return EBERR_ACTION_ASSIGNED;
      }
    }

    // Register the action
    actionhandler *handler = new actionhandler;
    handler->actionname = nameref;
    handler->aobject = aobject;
    globalhandlers.push_back(handler);
    return EBERR_SUCCESS;
  }
  else {
    // Not a valid user action
    cerr << "'" << actionname << "' is not a valid user action" << endl;
    return EBERR_ACTION_INVALID;
  }
}

void ActionManager::UnregisterLocalAction(string actionname, ActionObject *aobject) {
  for(unsigned int i=0; i<localhandlers.size();i++) {
    if(localhandlers[i]->aobject == aobject && localhandlers[i]->actionname == actionname) {
      actionhandler *handler = localhandlers[i];
      localhandlers.erase(localhandlers.begin() + i);
      delete handler;
      break;
    }
  }
}

void ActionManager::UnregisterGlobalAction(string actionname, ActionObject *aobject) {
  for(unsigned int i=0; i<globalhandlers.size();i++) {
    if(globalhandlers[i]->aobject == aobject && globalhandlers[i]->actionname == actionname) {
      actionhandler *handler = globalhandlers[i];
      globalhandlers.erase(globalhandlers.begin() + i);
      delete handler;
      break;
    }
  }
}

void ActionManager::UnregisterObjectActions(ActionObject *aobject) {
  actionhandler *handler;

  for(unsigned i=0; i<localhandlers.size();) {
    if(localhandlers[i]->aobject == aobject) {
      handler = localhandlers[i];
      localhandlers.erase(localhandlers.begin() + i);
      delete handler;
    }
    else
      i++;
  }
  for(unsigned i=0; i<globalhandlers.size();) {
    if(globalhandlers[i]->aobject == aobject) {
      handler = globalhandlers[i];
      globalhandlers.erase(globalhandlers.begin() + i);
      delete handler;
    }
    else
      i++;
  }
}

int ActionManager::ExecuteAction(string action) {
  // Find one action to execute, and execute it

  if(action.length()) {
    // First look for a local action
    for(unsigned int i=0; i<localhandlers.size(); i++) {
      if(localhandlers[i]->actionname == action) {
        localhandlers[i]->aobject->Execute();
        return 1;
      }
    }
    // Didn't find a local action, look for a global action
    for(unsigned int i=0; i<globalhandlers.size(); i++) {
      if(globalhandlers[i]->actionname == action) {
        globalhandlers[i]->aobject->Execute();
        return 2;
      }
    }
  }
  // No action found
  return 0;
}

string GetUserAction(string actionname) {
  // Get a pointer into the array of user action name strings
  // (verifies the name and saves storing a copy of the name at the same time)
  for(int i=0; useractions[i] != NULL; i++) {
    if(useractions[i] == actionname) {
      return useractions[i];
    }
  }
  return "";
}
