/***************************************************************************
                          customobject.cpp  -  eboxy custom object
                             -------------------
    begin                : Sun Sep 8 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <cstring>
#include <cstdlib>

#include "customobject.h"
#include "membermethodhandler.h"
#include "memberfunctionproperty.h"
#include "functionproperty.h"
#include "../errors.h"
#include "utils.h"

using namespace std;

CustomObject::CustomObject(void) {
  this->type = "CustomObject"; // If this is changed, need to change dynamicobject::ClearEventHandlers
  this->defaultproperty = NULL;

  MemberFunctionProperty<CustomObject> *mp;
  mp = new MemberFunctionProperty<CustomObject>("name", this, &CustomObject::pget_Name, NULL, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<CustomObject>("type", this, &CustomObject::pget_Type, NULL, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<CustomObject>("propertycount", this, &CustomObject::pget_PropertyCount, NULL, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<CustomObject>("methodcount", this, &CustomObject::pget_MethodCount, NULL, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<CustomObject>("eventcount", this, &CustomObject::pget_EventCount, NULL, false);
  AddProperty(mp);

  MemberMethodHandler<CustomObject> *mh;
  mh = new MemberMethodHandler<CustomObject>("getproperty", this, 1, &CustomObject::m_GetProperty);
  AddMethod(mh);
  mh = new MemberMethodHandler<CustomObject>("getmethod", this, 1, &CustomObject::m_GetMethod);
  AddMethod(mh);
  mh = new MemberMethodHandler<CustomObject>("getevent", this, 1, &CustomObject::m_GetEvent);
  AddMethod(mh);
}

CustomObject::~CustomObject(void) {
  while(properties.size() > 0) {
    delete properties.back();
    properties.pop_back();
  }
}

int CustomObject::AddProperty(const char *name, propertygetfn getter, propertysetfn setter) {
  if(CheckMember(name))
    return EBERR_MEMBER_EXISTS;

  FunctionProperty *prop = new FunctionProperty(name, getter, setter);
  properties.push_back(prop);

  return EBERR_SUCCESS;
}

int CustomObject::AddProperty(CustomProperty *prop) {
  if(CheckMember(prop->GetName()))
    return EBERR_MEMBER_EXISTS;

  properties.push_back(prop);

  return EBERR_SUCCESS;
}

int CustomObject::AddMethod(MethodHandler *mh) {
  if(CheckMember(mh->GetName()))
    return EBERR_MEMBER_EXISTS;

  methodhandlers.push_back(mh);

  return EBERR_SUCCESS;
}

Variant CustomObject::GetPropertyValue(const char *property) {
  CustomProperty *prop = FindProperty(property);
  if(prop)
    return prop->GetValue();
  else {
    cerr << "Object " << GetName() << " does not have a property " << property << endl;
    return VARNULL;
  }
}

int CustomObject::SetPropertyValue(const char *property, Variant value) {
  CustomProperty *prop = FindProperty(property);
  if(prop)
    return prop->SetValue(value);
  else {
    cerr << "Object " << GetName() << " does not have a property " << property << endl;
    return EBERR_MEMBER_NOT_FOUND;
  }
}

CustomProperty *CustomObject::FindProperty(const char *name) {
  for(unsigned int i=0; i<properties.size(); i++) {
    if(!strcasecmp(properties[i]->GetName(), name))
      return properties[i];
  }
  return NULL;
}

CustomProperty *CustomObject::GetProperty(int index) {
  if(index < properties.size() && index >= 0)
    return properties[index];
  else
    return NULL;
}

int CustomObject::GetPropertyCount(void) {
  // of course, this DOES NOT count hacked in properties in child classes
  // (which is why they shouldn't be done that way)
  return properties.size();
}

const char *CustomObject::GetDefaultProperty(void) {
  if(defaultproperty)
    return defaultproperty->GetName();
  else
    return NULL;
}

bool CustomObject::HasProperty(const char *property) {
  return ((this->FindProperty(property)) != NULL);
}

int CustomObject::FireEvent(const char *event) {
  return DynamicObject::FireEvent(event);
}

int CustomObject::FireEventWait(const char *event) {
  return DynamicObject::FireEventWait(event);
}

bool CustomObject::CheckMember(const char *name) {
  if(HasProperty(name)) {
    cerr << "Object " << GetName() << " already has a property " << name << endl;
    return true;
  }
  else if(HasMethod(name)) {
    cerr << "Object " << GetName() << " already has a method " << name << endl;
    return true;
  }
  return false;
}

/* Property get/set functions */

// -- name

Variant CustomObject::pget_Name(void) {
  return anytovariant(GetName());
}

// -- type

Variant CustomObject::pget_Type(void) {
  return anytovariant(GetType());
}

// -- propertycount

Variant CustomObject::pget_PropertyCount(void) {
  return anytovariant(GetPropertyCount());
}

// -- methodcount

Variant CustomObject::pget_MethodCount(void) {
  return anytovariant(GetMethodCount());
}

// -- eventcount

Variant CustomObject::pget_EventCount(void) {
  return anytovariant(GetEventCount());
}

/* Method functions */

Variant CustomObject::m_GetProperty(int numargs, Variant args[]) {
  CustomProperty *prop = GetProperty(args[0]);
  if(prop)
    return anytovariant(prop->GetName());
  else
    return VARNULL;
}

Variant CustomObject::m_GetMethod(int numargs, Variant args[]) {
  MethodHandler *mthd = GetMethod(args[0]);
  if(mthd)
    return anytovariant(mthd->GetName());
  else
    return VARNULL;
}

Variant CustomObject::m_GetEvent(int numargs, Variant args[]) {
  const char *eventname = GetEvent(args[0]);
  if(eventname)
    return anytovariant(eventname);
  else
    return VARNULL;
}
