/***************************************************************************
           dynamicobject.cpp  -  parent object class supporting events
                             -------------------
    begin                : Sun Sep 22 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>

#include "dynamicobject.h"
#include "functioneventhandler.h"
#include "scripteventhandler.h"
#include "../errors.h"

using namespace std;

extern const char *TYPESTR_PLUGIN;

DynamicObject::DynamicObject(void) {
  this->type = "_DynamicObject";
  findable = true;
  container = NULL;
  defaultevent = -1;
}

DynamicObject::~DynamicObject(void) {
  ClearEventHandlers();
  ClearMethodHandlers();
  if(this->container)
    container->Notify(this);
}

EventHandler *DynamicObject::RegisterEventHandler(const char *event, Object *sink, int (*eventfunction)(const char *)) {
  EventHandler *handler = new FunctionEventHandler(event, this, sink, eventfunction);
  eventhandlers.push_back(handler);
  return handler;
}

EventHandler *DynamicObject::RegisterEventHandler(const char *event, ScriptBase *script) {
  EventHandler *handler = new ScriptEventHandler(event, this, NULL, script);
  eventhandlers.push_back(handler);
  return handler;
}

ExternalMethodHandler *DynamicObject::RegisterMethod(const char *method, Object *owner, int numargs, methodfn methodfunc) {
  ExternalMethodHandler *handler = new ExternalMethodHandler(method, this, numargs, methodfunc);
  methodhandlers.push_back(handler);
  return handler;
}

void DynamicObject::UnregisterEventHandler(EventHandler *handler) {
  for(unsigned int i=0;i<eventhandlers.size(); i++) {
    if(eventhandlers[i] == handler) {
      eventhandlers.erase(eventhandlers.begin() + i);
      delete handler;
      break;
    }
  }
}

int DynamicObject::UnregisterEventHandler(const char *event, Object *sink) {
  // Unregister all event handlers for the specified event registered by the specified sink
  int unregcount = 0;

  for(unsigned int i=0;i<eventhandlers.size();) {
    EventHandler *handler = eventhandlers[i];
    if(handler->GetSink() == sink && !strcasecmp(handler->GetName(), event)) {
      eventhandlers.erase(eventhandlers.begin() + i);
      delete handler;
      unregcount++;
    }
    else
      i++;
  }

  if(unregcount == 0)
    return EBERR_EVENT_NOT_HANDLED;
  else
    return EBERR_SUCCESS;
}

void DynamicObject::UnregisterMethod(MethodHandler *handler) {
  for(unsigned int i=0;i<methodhandlers.size(); i++) {
    if(methodhandlers[i] == handler) {
      methodhandlers.erase(methodhandlers.begin() + i);
      delete handler;
      break;
    }
  }
}

int DynamicObject::FireEvent(const char *event) {
  for(unsigned int i=0;i<eventhandlers.size(); i++) {
    if(!strcasecmp(eventhandlers[i]->GetName(), event)) {
      eventhandlers[i]->Fire(true);
    }
  }
  return 0;
}

int DynamicObject::FireEventWait(const char *event) {
  for(unsigned int i=0;i<eventhandlers.size(); i++) {
    if(!strcasecmp(eventhandlers[i]->GetName(), event))
      eventhandlers[i]->Fire(false);
  }
  return 0;
}

bool DynamicObject::EventHandled(const char *event) {
  for(unsigned int i=0;i<eventhandlers.size(); i++) {
    if(!strcasecmp(eventhandlers[i]->GetName(), event))
      return true;
  }
  return false;
}

Variant DynamicObject::MethodCall(const char *method, int numargs, Variant args[]) {
  for(unsigned int i=0;i<methodhandlers.size(); i++) {
    if(!strcasecmp(methodhandlers[i]->GetName(), method))
      return methodhandlers[i]->Call(numargs, args);
  }
  // Missing method
  cerr << "Object " << GetName() << " does not have a method " << method << endl;
  return VARNULL;
}

Variant DynamicObject::GetPropertyValue(const char *property) {
  if(!strcasecmp(property, "name"))
    return anytovariant(GetName());
  else if(!strcasecmp(property, "type"))
    return anytovariant(GetType());
  else {
    cerr << "Object " << GetName() << " does not have a property " << property << endl;
    return VARNULL;
  }
}

int DynamicObject::SetPropertyValue(const char *property, Variant value) {
  if(!strcasecmp(property, "name")) {
    cerr << "Property \'name\' is read-only" << endl;
    return -1;
  }
  else if(!strcasecmp(property, "type")) {
    cerr << "Property \'type\' is read-only" << endl;
    return -1;
  }
  else {
    cerr << "Object " << GetName() << " does not have a property " << property << endl;
    return -1;
  }
}

bool DynamicObject::HasProperty(const char *property) {
  if(!strcasecmp("name", property) || !strcasecmp("type", property))
    return true;
  else
    return false;
}

bool DynamicObject::HasMethod(const char *method) {
  for(unsigned int i=0;i<methodhandlers.size(); i++) {
    if(!strcasecmp(methodhandlers[i]->GetName(), method))
      return true;
  }
  return false;
}

MethodHandler *DynamicObject::GetMethod(int index) {
  if(index < methodhandlers.size() && index >= 0)
    return methodhandlers[index];
  else
    return NULL;
}

int DynamicObject::GetMethodCount(void) {
  // of course, this DOES NOT count hacked in properties in child classes
  // (which is why they shouldn't be done that way)
  return methodhandlers.size();
}

void DynamicObject::ClearEventHandlers(void) {
  // Remove all attached event handlers
  while(eventhandlers.size() > 0) {
    EventHandler *handler = eventhandlers.back();
    eventhandlers.pop_back();
    if(handler->GetSource() != NULL) {
      // Check for plugin-created objects
      if(!strcmp(handler->GetSource()->GetType(), "CustomObject"))
        handler->GetSource()->GetOwner()->RemoveOwnedObject(handler);
    }
    if(handler->GetSink() != NULL) {
      // Check for other plugin-owned events
      // FIXME: I don't like these type checks... is there a better way to do this?
      DynamicObject *dobj = dynamic_cast<DynamicObject *>(handler->GetSink());
      if(!strcmp(dobj->GetType() , TYPESTR_PLUGIN)) {
        // The handler belongs to a plugin, so we need to tell the plugin object
        // to stop holding onto a pointer which will shortly be invalid
        ((OwnerObject *)(handler->GetSink()))->RemoveOwnedObject(handler);
      }
    }
    delete handler;
  }
}

void DynamicObject::ClearMethodHandlers(void) {
  // Remove all attached method handlers
  while(methodhandlers.size() > 0) {
    MethodHandler *handler = methodhandlers.back();
    methodhandlers.pop_back();
    if(handler->GetOwner() != NULL) {
      if(!strcmp(handler->GetOwner()->GetType(), "CustomObject")) {
        if(!strcmp((dynamic_cast<DynamicObject *>(handler->GetOwner()->GetOwner()))->GetType(), TYPESTR_PLUGIN)) {
          // The handler belongs to an object, which belongs plugin, so we need to tell the plugin
          // to stop holding onto a pointer which will shortly be invalid
          ((OwnerObject *)(handler->GetOwner()->GetOwner()))->RemoveOwnedObject(handler);
        }
      }
    }
    delete handler;
  }
}

bool DynamicObject::IsFindable(void) {
  return findable;
}

void DynamicObject::SetContainer(NotifyObject *container) {
  if(this->container != NULL && container != NULL) {
    cerr << "DynamicObject::SetContainer - container already set!" << endl;
  }
  else
    this->container = container;
}

void DynamicObject::AddEvent(const char *eventname) {
  events.Add(eventname);
}

int DynamicObject::GetEventCount(void) {
  return events.size();
}

const char *DynamicObject::GetEvent(int index) {
  if(index >= 0 && index < events.size())
    return events[index];
  else
    return NULL;
}

const char *DynamicObject::GetDefaultEvent(void) {
  if(defaultevent > -1)
    return events[defaultevent];
  else
    return NULL;
}

