/***************************************************************************
                  pagebase.cpp  -  base class for eboxy pages
                             -------------------
    begin                : Mon Sep 22 2003
    copyright            : (C) 2003 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>

#include "utils.h"
#include "timerbase.h"
#include "pagebase.h"
#include "memberfunctionproperty.h"
#include "membermethodhandler.h"
#include "resolverbase.h"
#include "resourcemanagerbase.h"
#include "actionmanager.h"
#include "../errors.h"
 
using namespace std;

extern const char *TYPESTR_PAGE;
extern const char *TYPESTR_TIMER;
extern ResolverBase *resolver;
extern ResourceManagerBase *resourcemanager;
extern ActionManager *actionmanager;

PageBase::PageBase(const char *name, int x, int y, int w, int h)
: Object(name), WidgetBase(x, y) {
  this->w = w;
  this->h = h;
  this->background = NULL;
  this->type = TYPESTR_PAGE;

  MemberFunctionProperty<PageBase> *mp;
  mp = new MemberFunctionProperty<PageBase>("widgetcount", this, &PageBase::pget_WidgetCount, NULL, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<PageBase>("background", this, &PageBase::pget_Background, &PageBase::pset_Background, false);
  AddProperty(mp);

  MemberMethodHandler<PageBase> *mh;
  mh = new MemberMethodHandler<PageBase>("getwidget", this, 1, &PageBase::m_GetWidget);
  AddMethod(mh);
  mh = new MemberMethodHandler<PageBase>("addwidget", this, 1, &PageBase::m_AddWidget);
  AddMethod(mh);
  mh = new MemberMethodHandler<PageBase>("removewidget", this, 1, &PageBase::m_RemoveWidget);
  AddMethod(mh);

  focusedindex = -1;
  
  AddEvent("OnLoad");
  AddEvent("OnShow");
  AddEvent("OnUnload");
}

PageBase::~PageBase(void) {
  while(widgets.size() > 0) {
    WidgetBase *wb = widgets.back();
    widgets.pop_back();
    delete wb;
  }
  while(actionhandlers.size() > 0) {
    ActionHandler *handler = actionhandlers.back();
    actionhandlers.pop_back();
    delete handler;
  }
}

void PageBase::SetBackground(ImageResource *image) {
  background = image;
}

void PageBase::FocusNextWidget(void) {
  int check;

  if(widgets.size() == 0) {
    // No widgets
    focusedindex = -1;
    return;
  }

  // Unfocus the currently focused widget, if any
  if(focusedindex >= 0)
    widgets[focusedindex]->LostFocus();

  // Move to next widget
  focusedindex++;
  if(focusedindex > (widgets.size() - 1))
    focusedindex = 0;

  // Now we need to make sure the next widget is focusable, if not find the next focusable one
  check = focusedindex;
  while(focusedindex > -1 && !((widgets[focusedindex]->GetTabStop()) && widgets[focusedindex]->IsEnabled())) {
    focusedindex++;
    if(focusedindex > (widgets.size() - 1))
      focusedindex = 0;
    if(focusedindex == check)  // no more left, can't set focus
      focusedindex = -1;
  }

  // Finally, give focus to the newly focused widget, if any
  if(focusedindex >= 0)
    widgets[focusedindex]->GotFocus();
}

void PageBase::FocusPreviousWidget(void) {
  int check;

  if(widgets.size() == 0) {
    // No widgets
    focusedindex = -1;
    return;
  }

  // Unfocus the currently focused widget, if any
  if(focusedindex >= 0)
    widgets[focusedindex]->LostFocus();

  // Move to previous widget
  focusedindex--;
  if(focusedindex < 0)
    focusedindex = widgets.size() - 1;

  // Now we need to make sure the previous widget is focusable, if not find the previous focusable one
  check = focusedindex;
  while(focusedindex > -1 && !((widgets[focusedindex]->GetTabStop()) && widgets[focusedindex]->IsEnabled())) {
    focusedindex--;
    if(focusedindex < 0)
      focusedindex = widgets.size() - 1;
    if(focusedindex == check)  // no more left, can't set focus
      focusedindex = -1;
  }

  // Finally, give focus to the newly focused widget, if any
  if(focusedindex >= 0)
    widgets[focusedindex]->GotFocus();
}

void PageBase::SetFocus(WidgetBase *widget) {
  // Find widget in array
  for(unsigned int i=0; i<widgets.size(); i++) {
    if(widgets[i]==widget) {
      if(i != focusedindex && widgets[i]->GetTabStop() && widgets[i]->IsEnabled() ) {
        // Give widget focus
        widgets[focusedindex]->LostFocus();
        focusedindex = i;
        widgets[focusedindex]->GotFocus();
      }
      break;
    }
  }
}

WidgetBase *PageBase::GetFocusedWidget(void) {
  if(focusedindex >= 0)
    return widgets[focusedindex];
  else
    return NULL;
}

DynamicObject *PageBase::FindObject(const char *name, const char *type) {
  WidgetBase *widget = FindWidget(name, type);
  if(widget)
    return widget;
  else
    return ObjectContainer::FindObject(name, type);
}

WidgetBase *PageBase::FindWidget(const char *name, const char *type) {
  for(unsigned int i=0; i<widgets.size(); i++) {
    if(!strcasecmp(widgets[i]->GetName(), name) && (type == NULL || (!strcasecmp(widgets[i]->GetType(), type))))
      return widgets[i];
  }
  return NULL;
}

void PageBase::DoOnShow(void) {
  // Called by switchPage when the page has done switching
  this->FireEventWait("OnShow");

  // Set focused widget
  if(focusedindex == -1) {
    FocusNextWidget();
  }
  else {
    UpdateFocus();
  }

  // Start any auto-starting timers
  for(unsigned int i=0;i<containedobjects.size();i++) {
    DynamicObject *object = containedobjects[i];
    if(!strcmp(object->GetType(), TYPESTR_TIMER)) {
      // Start the timer
      (dynamic_cast<TimerObjectBase *>(object))->AutoStart();
    }
  }
}

void PageBase::DoInstantiate(void) {
  // Register actions for actionhandlers
  for(unsigned int i=0;i<actionhandlers.size();i++) {
    std::string action = actionhandlers[i]->GetAction();
    int rc = actionmanager->RegisterLocalAction(action, actionhandlers[i]);
    if(rc == EBERR_ACTION_INVALID) {
      cerr << "Warning: action \"" << action << "\" is not a valid user action" << endl;
    }
    else if(rc == EBERR_ACTION_ASSIGNED) {
      cerr << "Warning: action \"" << action << "\" is assigned to more than one widget in this context" << endl;
    }
  }
}

void PageBase::DoUninstantiate(void) {
  // Unregister actions for actionhandlers
  for(unsigned int i=0;i<actionhandlers.size();i++) {
    actionmanager->UnregisterLocalAction(actionhandlers[i]->GetAction(), actionhandlers[i]);
  }
}

void PageBase::AddActionHandler(ActionHandler *handler) {
  actionhandlers.push_back(handler);
}

void PageBase::RemoveActionHandler(ActionHandler *handler) {
  for(unsigned int i=0;i<actionhandlers.size();i++) {
    if(handler == actionhandlers[i]) {
      actionhandlers.erase(actionhandlers.begin() + i);
      break;
    }
  } 
}


/* Property get/set functions */

//-- widgetcount

Variant PageBase::pget_WidgetCount(void) {
  return anytovariant(int(widgets.size()));
}

Variant PageBase::pget_Background(void) {
  if(background)
    return anytovariant(background->GetFilename());
  else
    return VARNULL;
}

int PageBase::pset_Background(Variant value) {
  ImageResource *image = resourcemanager->LoadImageResource(value, false);
  if(image) {
    SetBackground(image);
    return EBERR_SUCCESS;
  }
  else
    return EBERR_FILE_ERROR;
}

/* Method functions */

Variant PageBase::m_AddWidget(int numargs, Variant args[]) {
  string widgetname = args[0];
  WidgetBase *widget = dynamic_cast<WidgetBase *>(resolver->Resolve(widgetname.c_str(), RSCOPE_TEMP));

  if(widget)
    AddWidget(widget);
  else
    cerr << "Page.AddWidget: Unable to find widget named " << widgetname << " in constructed widgets" << endl;

  return VARNULL;
}

Variant PageBase::m_RemoveWidget(int numargs, Variant args[]) {
  string widgetname = args[0];
  WidgetBase *widget = dynamic_cast<WidgetBase *>(FindWidget(widgetname.c_str(), NULL));
  if(widget) {
    if(resolver->Resolve(widgetname.c_str(), RSCOPE_TEMP) != widget)
      cerr << "Page.RemoveWidget: widget " << widgetname << " cannot be removed - was not constructed" << endl;
    else
      RemoveWidget(widget);
  }
  else
    cerr << "Page.RemoveWidget: no widget named " << widgetname << " exists on page " << GetName() << endl;

  return VARNULL;
}

Variant PageBase::m_GetWidget(int numargs, Variant args[]) {
  int index = args[0];
  if(index >= 0 && index < widgets.size())
    return anytovariant(widgets[index]->GetName());
  else {
    cerr << "Page.GetWidget: index " << index << " is out of bounds" << endl;
    return VARNULL;
  }
}
