/***************************************************************************
          scrollbarwidgetbase.cpp  -  base class for scrollbar widgets
                             -------------------
    begin                : Mon Nov 17 2003
    copyright            : (C) 2003 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "utils.h"
#include "resourcemanagerbase.h"
#include "memberfunctionproperty.h"
#include "../errors.h"

#include "scrollbarwidgetbase.h"

using namespace std;

extern const char *TYPESTR_SCROLLBAR;
extern ResourceManagerBase *resourcemanager;

ScrollBarWidgetBase::ScrollBarWidgetBase(const char *name, int x, int y, int w, int h)
: WidgetBase(x, y) {
  this->w = w;
  this->h = h;
  // Set defaults
  tabstop = true;
  value = 0;
  pagestep = 10;
  reversed = false;
  minimum = 0;
  maximum = 100;
  background = NULL;
  knob = NULL;
  knobfocus = NULL;
  backgroundstyle = IMAGESTYLE_NORMAL;

  this->type = TYPESTR_SCROLLBAR;

  MemberFunctionProperty<ScrollBarWidgetBase> *mp;
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("background", this, &ScrollBarWidgetBase::pget_Background, &ScrollBarWidgetBase::pset_Background, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("backgroundstyle", this, &ScrollBarWidgetBase::pget_BackgroundStyle, &ScrollBarWidgetBase::pset_BackgroundStyle, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("knob", this, &ScrollBarWidgetBase::pget_Knob, &ScrollBarWidgetBase::pset_Knob, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("knobfocus", this, &ScrollBarWidgetBase::pget_KnobFocus, &ScrollBarWidgetBase::pset_KnobFocus, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("value", this, &ScrollBarWidgetBase::pget_Value, &ScrollBarWidgetBase::pset_Value, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("maximum", this, &ScrollBarWidgetBase::pget_Maximum, &ScrollBarWidgetBase::pset_Maximum, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("minimum", this, &ScrollBarWidgetBase::pget_Minimum, &ScrollBarWidgetBase::pset_Minimum, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("pagestep", this, &ScrollBarWidgetBase::pget_PageStep, &ScrollBarWidgetBase::pset_PageStep, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<ScrollBarWidgetBase>("reversed", this, &ScrollBarWidgetBase::pget_Reversed, &ScrollBarWidgetBase::pset_Reversed, false);
  AddProperty(mp);

  FindProperty("width")->SetRequired(true);
  FindProperty("height")->SetRequired(true);

  AddEvent("OnChange");
  AddEvent("OnGotFocus");
  AddEvent("OnLostFocus");
}

void ScrollBarWidgetBase::SetPageStep(int pagestep) {
  this->pagestep = pagestep;
}

void ScrollBarWidgetBase::SetReversed(bool reversed) {
  this->reversed = reversed;
}

void ScrollBarWidgetBase::Increment(void) {
  int newvalue = value + 1;

  if(newvalue <= maximum)
    SetValue(newvalue);
  else
    SetValue(maximum);
}

void ScrollBarWidgetBase::Decrement(void) {
  int newvalue = value - 1;

  if(newvalue >= minimum)
    SetValue(newvalue);
  else
    SetValue(minimum);
}

void ScrollBarWidgetBase::IncrementPage(void) {
  int newvalue = value + pagestep;

  if(newvalue <= maximum)
    SetValue(newvalue);
  else
    SetValue(maximum);
}

void ScrollBarWidgetBase::DecrementPage(void) {
  int newvalue = value - pagestep;

  if(newvalue >= minimum)
    SetValue(newvalue);
  else
    SetValue(minimum);
}

void ScrollBarWidgetBase::GotoStart(void) {
  SetValue(minimum);
}

void ScrollBarWidgetBase::GotoEnd(void) {
  SetValue(maximum);
}

bool ScrollBarWidgetBase::StandardAction(int actiontype) {
  switch(actiontype) {
    case STDACTION_ITEMPREV:
      Decrement();
      return true;
      break;
    case STDACTION_ITEMNEXT:
      Increment();
      return true;
      break;
    case STDACTION_ITEMPAGEPREV:
      DecrementPage();
      return true;
      break;
    case STDACTION_ITEMPAGENEXT:
      IncrementPage();
      return true;
      break;
    case STDACTION_ITEMFIRST:
      GotoStart();
      return true;
      break;
    case STDACTION_ITEMLAST:
      GotoEnd();
      return true;
      break;
    default:
      return false;
  }
}


/* Property get/set functions */

// -- background --

Variant ScrollBarWidgetBase::pget_Background(void) {
  if( background )
    return anytovariant(background->GetFilename());
  else
    return VARNULL;
}

int ScrollBarWidgetBase::pset_Background(Variant value) {
  ImageResource *image = resourcemanager->LoadImageResource(value, false);
  if(image) {
    SetBackgroundImage(image);
    return EBERR_SUCCESS;
  }
  else
    return EBERR_FILE_ERROR;
}

// -- backgroundstyle --

Variant ScrollBarWidgetBase::pget_BackgroundStyle(void) {
  return anytovariant(styleinttostr(backgroundstyle));
}

int ScrollBarWidgetBase::pset_BackgroundStyle(Variant value) {
  SetBackgroundStyle(stylestrtoint(value));
  return EBERR_SUCCESS;
}

// -- knob --

Variant ScrollBarWidgetBase::pget_Knob(void) {
  if( knob )
    return anytovariant(knob->GetFilename());
  else
    return VARNULL;
}

int ScrollBarWidgetBase::pset_Knob(Variant value) {
  ImageResource *image = resourcemanager->LoadImageResource(value, false);
  if(image) {
    SetKnobImage(image);
    return EBERR_SUCCESS;
  }
  else
    return EBERR_FILE_ERROR;
}

// -- knobfocus --

Variant ScrollBarWidgetBase::pget_KnobFocus(void) {
  if( knobfocus )
    return anytovariant(knobfocus->GetFilename());
  else
    return VARNULL;
}

int ScrollBarWidgetBase::pset_KnobFocus(Variant value) {
  ImageResource *image = resourcemanager->LoadImageResource(value, false);
  if(image) {
    SetKnobFocusImage(image);
    return EBERR_SUCCESS;
  }
  else
    return EBERR_FILE_ERROR;
}

// -- value --

Variant ScrollBarWidgetBase::pget_Value(void) {
  return anytovariant(this->value);
}

int ScrollBarWidgetBase::pset_Value(Variant value) {
  SetValue(value);
  return EBERR_SUCCESS;
}

// -- maximum --

Variant ScrollBarWidgetBase::pget_Maximum(void) {
  return anytovariant(maximum);
}

int ScrollBarWidgetBase::pset_Maximum(Variant value) {
  SetMaximum(value);
  return EBERR_SUCCESS;
}

// -- minimum --

Variant ScrollBarWidgetBase::pget_Minimum(void) {
  return anytovariant(minimum);
}

int ScrollBarWidgetBase::pset_Minimum(Variant value) {
  SetMinimum(value);
  return EBERR_SUCCESS;
}

// -- pagestep --

Variant ScrollBarWidgetBase::pget_PageStep(void) {
  return anytovariant(pagestep);
}

int ScrollBarWidgetBase::pset_PageStep(Variant value) {
  SetPageStep(value);
  return EBERR_SUCCESS;
}

// -- reversed --

Variant ScrollBarWidgetBase::pget_Reversed(void) {
  return anytovariant(reversed);
}

int ScrollBarWidgetBase::pset_Reversed(Variant value) {
  SetReversed(value);
  return EBERR_SUCCESS;
}
