/***************************************************************************
        textfieldwidgetbase.cpp  -  base class for textfield widgets
                             -------------------
    begin                : Fri Oct 3 2003
    copyright            : (C) 2003 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
 
#include "utils.h"
#include "../errors.h"

#include "memberfunctionproperty.h"
#include "resourcemanagerbase.h"
#include "textfieldwidgetbase.h"
#include "pagebase.h"
   
extern const char *TYPESTR_TEXTFIELD;
extern const int DEFAULT_FONTSIZE;

extern ResourceManagerBase *resourcemanager;

using namespace std;

// This array defines what characters are available using keypad (digit) text input
const char *digitchars[] = {
  "0 -+",
  "1.,:;\'\"",
  "abc2",
  "def3",
  "ghi4",
  "jkl5",
  "mno6",
  "pqrs7",
  "tuv8",
  "wxyz9",
  NULL  // end of list
};

TextFieldWidgetBase::TextFieldWidgetBase(const char *name, int x, int y, int w, int h, FontResource *font, eboxy_color fontcolor, int maxlength)
: Object(name), WidgetBase(x, y) {
  this->w = w;
  this->h = h;
  this->font = font;
  this->fontsize = DEFAULT_FONTSIZE;
  this->fontcolor = fontcolor;
  this->maxlength = maxlength;

  this->inpdigit = -1;
  this->inpdigitptr = NULL;

  // Set defaults
  this->text = "";
  this->validchars = "";

  this->tabstop = true;
  this->type = TYPESTR_TEXTFIELD;
  this->background = NULL;
  this->backgroundfocus = NULL;
  this->backgroundstyle = IMAGESTYLE_NORMAL;
  this->borderx = 0;
  this->bordery = 0;

  MemberFunctionProperty<TextFieldWidgetBase> *mp;
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("text", this, &TextFieldWidgetBase::pget_Text, &TextFieldWidgetBase::pset_Text, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("font", this, &TextFieldWidgetBase::pget_Font, &TextFieldWidgetBase::pset_Font, true);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("fontsize", this, &TextFieldWidgetBase::pget_FontSize, &TextFieldWidgetBase::pset_FontSize, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("fontcolor", this, &TextFieldWidgetBase::pget_FontColor, &TextFieldWidgetBase::pset_FontColor, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("background", this, &TextFieldWidgetBase::pget_Background, &TextFieldWidgetBase::pset_Background, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("backgroundfocus", this, &TextFieldWidgetBase::pget_BackgroundFocus, &TextFieldWidgetBase::pset_BackgroundFocus, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("backgroundstyle", this, &TextFieldWidgetBase::pget_BackgroundStyle, &TextFieldWidgetBase::pset_BackgroundStyle, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("maxlength", this, &TextFieldWidgetBase::pget_MaxLength, &TextFieldWidgetBase::pset_MaxLength, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("borderx", this, &TextFieldWidgetBase::pget_BorderX, &TextFieldWidgetBase::pset_BorderX, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("bordery", this, &TextFieldWidgetBase::pget_BorderY, &TextFieldWidgetBase::pset_BorderY, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("cursorpos", this, &TextFieldWidgetBase::pget_CursorPos, &TextFieldWidgetBase::pset_CursorPos, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<TextFieldWidgetBase>("validchars", this, &TextFieldWidgetBase::pget_ValidChars, &TextFieldWidgetBase::pset_ValidChars, false);
  AddProperty(mp);

  FindProperty("width")->SetRequired(true);
  FindProperty("height")->SetRequired(true);

  AddEvent("OnChange");
  AddEvent("OnGotFocus");
  AddEvent("OnLostFocus");
}

TextFieldWidgetBase::~TextFieldWidgetBase(void) {

}

void TextFieldWidgetBase::SetText(std::string str) {
  if(str.length() > maxlength)
    this->text = str.substr(0, maxlength);
  else
    this->text = str;
}

std::string TextFieldWidgetBase::GetText(void) {
  return text;
}

// FIXME not i18n compatible
void TextFieldWidgetBase::InputDigit(int chr) {
  timeval currtime = getSystemTime();

  if(chr == -1) {
    input_time.tv_sec = 0;
    input_time.tv_usec = 0;
  }
  else if(inpdigit == chr && compareTimes(&currtime, &input_time)) {
    // We are already in digit input mode and the same button was pressed within the time limit
    // In this case, we rotate through the available digits for this character
    bool was_valid = false;
    const char *lastptr = inpdigitptr;
    TextFieldInsertMode insmode = INSMODE_BACKOVER;
    while(!was_valid) {
      inpdigitptr++;
      if(inpdigitptr[0] == '\0')
        inpdigitptr = digitchars[inpdigit];  // back to start
      if(inpdigitptr == lastptr) {
        // That's it... there's only one valid char, so just insert it again
        insmode = INSMODE_INSERT;
      }
      was_valid = SendChar(inpdigitptr[0], insmode);
    }
    // Reset timer
    input_time = getSystemTime();
    input_time.tv_usec += TEXTINPUT_DELAY_TIME;
  }
  else {
    // New key
    inpdigit = chr;
    if(inpdigit != -1) {
      inpdigitptr = digitchars[inpdigit];  // back to start
      bool was_valid = false;
      while(true) {
        was_valid = SendChar(inpdigitptr[0], INSMODE_INSERT);
        if(was_valid)
          break;
        else
          inpdigitptr++;

        if(inpdigitptr[0] == '\0')
          break; // No valid characters
      }
      if(inpdigitptr[0] != '\0') {
        // Start timer
        input_time = getSystemTime();
        input_time.tv_usec += TEXTINPUT_DELAY_TIME;
      }
    }
  }
}

bool TextFieldWidgetBase::StandardAction(int actiontype) {
  switch(actiontype) {
    case STDACTION_ACCEPT:
      FindParentPage()->FocusNextWidget();
      return true;
      break;
    case STDACTION_ITEMPREV:
      SetCursorPos(GetCursorPos() - 1);
      return true;
      break;
    case STDACTION_ITEMNEXT:
      SetCursorPos(GetCursorPos() + 1);
      return true;
      break;
    case STDACTION_ITEMFIRST:
      SetCursorPos(0);
      return true;
      break;
    case STDACTION_ITEMLAST:
      SetCursorPos(GetText().length());
      return true;
      break;
    default:
      return false;
  }
}


/* Property get/set functions */

// -- text --

Variant TextFieldWidgetBase::pget_Text(void) {
  return anytovariant(text);
}

int TextFieldWidgetBase::pset_Text(Variant value) {
  SetText(value);
  return EBERR_SUCCESS;
}

// -- font --

Variant TextFieldWidgetBase::pget_Font(void) {
  return anytovariant(font->GetFilename());
}

int TextFieldWidgetBase::pset_Font(Variant value) {
  FontResource *font = resourcemanager->LoadFontResource(value, fontsize);
  if(font) {
    SetFont(font);
    return EBERR_SUCCESS;
  }
  else
    return EBERR_FILE_ERROR;
}

// -- fontsize --

Variant TextFieldWidgetBase::pget_FontSize(void) {
  return anytovariant(fontsize);
}

int TextFieldWidgetBase::pset_FontSize(Variant value) {
  fontsize = value;
  if(font)
    return pset_Font(anytovariant(font->GetFilename()));
  else
    return EBERR_SUCCESS;
}

// -- fontcolor --

Variant TextFieldWidgetBase::pget_FontColor(void) {
  return anytovariant(colortostr( fontcolor ));
}

int TextFieldWidgetBase::pset_FontColor(Variant value) {
  SetFontColor(strtocolor(value));
  return EBERR_SUCCESS;
}

// -- background --

Variant TextFieldWidgetBase::pget_Background(void) {
  if( background )
    return anytovariant(background->GetFilename());
  else
    return VARNULL;
}

int TextFieldWidgetBase::pset_Background(Variant value) {
  ImageResource *image = resourcemanager->LoadImageResource(value, false);
  if(image) {
    SetBackgroundImage(image);
    return EBERR_SUCCESS;
  }
  else
    return EBERR_FILE_ERROR;
}

// -- backgroundfocus --

Variant TextFieldWidgetBase::pget_BackgroundFocus(void) {
  if( backgroundfocus )
    return anytovariant(backgroundfocus->GetFilename());
  else
    return VARNULL;
}

int TextFieldWidgetBase::pset_BackgroundFocus(Variant value) {
  ImageResource *image = resourcemanager->LoadImageResource(value, false);
  if(image) {
    SetBackgroundFocusImage(image);
    return EBERR_SUCCESS;
  }
  else
    return EBERR_FILE_ERROR;
}

// -- backgroundstyle --

Variant TextFieldWidgetBase::pget_BackgroundStyle(void) {
  return anytovariant(styleinttostr( backgroundstyle ));
}

int TextFieldWidgetBase::pset_BackgroundStyle(Variant value) {
  SetBackgroundStyle(stylestrtoint(value));
  return EBERR_SUCCESS;
}

// -- maxlength --

Variant TextFieldWidgetBase::pget_MaxLength(void) {
  return anytovariant(maxlength);
}

int TextFieldWidgetBase::pset_MaxLength(Variant value) {
  if(!IsInstantiated()) {
    maxlength = value;
    return EBERR_SUCCESS;
  }
  else {
    cerr << "SetPropertyValue: Property maxlength is read-only" << endl;
    return EBERR_BAD_PROPERTY_VALUE;
  }
}

// -- borderx --

Variant TextFieldWidgetBase::pget_BorderX(void) {
  return anytovariant( borderx );
}

int TextFieldWidgetBase::pset_BorderX(Variant value) {
  SetBorderX(value);
  return EBERR_SUCCESS;
}

// -- bordery --

Variant TextFieldWidgetBase::pget_BorderY(void) {
  return anytovariant( bordery );
}

int TextFieldWidgetBase::pset_BorderY(Variant value) {
  SetBorderY(value);
  return EBERR_SUCCESS;
}

// -- cursorpos --

Variant TextFieldWidgetBase::pget_CursorPos(void) {
  return anytovariant(GetCursorPos());
}

int TextFieldWidgetBase::pset_CursorPos(Variant value) {
  SetCursorPos(value);
  return EBERR_SUCCESS;
}

// -- validchars --

Variant TextFieldWidgetBase::pget_ValidChars(void) {
  return anytovariant(validchars);
}

int TextFieldWidgetBase::pset_ValidChars(Variant value) {
  SetValidChars(value);
  return EBERR_SUCCESS;
}


timeval getSystemTime(void) {
  timeval t;
  gettimeofday(&t, NULL);
  return t;
}

bool compareTimes(timeval *t1, timeval *t2) {
  timeval t2a;

  t2a.tv_usec = t2->tv_usec;
  t2a.tv_sec = t2->tv_sec;

  /* Perform the carry for the later subtraction by updating y. */
  if (t1->tv_usec < t2a.tv_usec) {
    int nsec = (t2a.tv_usec - t1->tv_usec) / 1000000 + 1;
    t2a.tv_usec -= 1000000 * nsec;
    t2a.tv_sec += nsec;
  }
  if (t1->tv_usec - t2a.tv_usec > 1000000) {
    int nsec = (t1->tv_usec - t2a.tv_usec) / 1000000;
    t2a.tv_usec += 1000000 * nsec;
    t2a.tv_sec -= nsec;
  }

  return t1->tv_sec < t2a.tv_sec;
}
