/***************************************************************************
                          utils.cpp  -  eboxy general utility functions
                             -------------------
    begin                : Thu Aug 15 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstring>
#include <cctype>
#include <iostream>
#include <cstdlib>
#include <cstdio>

#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <limits.h>
#include <pwd.h>
#include <errno.h>
#include <regex.h>

#include "variant.h"

#include "utils.h"

using namespace std;

// Allocates space for a new string using malloc
char *newstring(int length) {
  // We use length+1 here to allow for the null terminator
  char *ret = (char *)malloc(length+1);
  if(ret) {
    ret[length] = '\0'; // for sanity's sake
    return ret;
  }
  else {
    // You'd hope this would never happen...
    cerr << "malloc("<< length+1 << ") failed" << endl;
    return NULL;
  }
}

// Trim off the whitespace at the start and end of the string
char *trimString(const char *string) {
  char *newStart;
  const char *p;
  int lastpos;

  // Find the first non-whitespace character
  p = string;
  for(unsigned int i=0; i < strlen(string); i++) {
    if(isspace(string[i]))
      p++;
    else
      break;
  }

  // Zero out all whitespace at the end of the string
  newStart = strdup(p);
  lastpos = strlen(newStart) - 1;
  for(int i = lastpos; i > 0; i--) {
    if(isspace(newStart[i]))
      newStart[i] = '\0';
    else
      break;
  }

  return newStart;
}

// Convert an HTML colour string (#rrggbb) into an eboxy_color structure
eboxy_color strtocolor(string colorstr) {
  eboxy_color color;
  int h;

  if(colorstr != "" && colorstr[0] == '#') {
    sscanf(colorstr.c_str()+1, "%X", &h);
    color.r = (h & 0xFF0000) >> 16;
    color.g = (h & 0xFF00) >> 8;
    color.b = (h & 0xFF);
  }
  else {
    color.r = 0;
    color.g = 0;
    color.b = 0;
  }
  return color;
}

// Convert an eboxy_color structure into an HTML colour string (#rrggbb)
string colortostr(eboxy_color color) {
  char colorstr[10];
  snprintf(colorstr, 8, "#%02X%02X%02X", color.r, color.g, color.b);
  return anytovariant(colorstr);
}

// Extracts the path component of a path+filename
std::string getFilePath(std::string &filename) {
  unsigned int pos = filename.rfind("/");
  if(pos > 0)
    return filename.substr(0, pos);
  else
    return "/";
}

// Expands any environment variables in the specified path
string expandPath(string filename) {
  if(filename == "")
    return "";

  string buf;
  string expandbuf;
  bool expanding = false;
  unsigned int i=0, j=0, k=0;
  int expandstart = 0;

  while(true) {
    if(expanding) {
      if(i >= filename.length() ||
          !(isalnum(filename[i]) || filename[i] == '_')) {
        // End of variable, try to dereference it
        if(filename[expandstart] == '$') {
          if(k==0 && i < filename.length() && filename[i] == '$') {
            // Special variable (process ID)
            char *pidstr = newstring(10);
            snprintf(pidstr, 10, "%d", getpid());
            int length = strlen(pidstr);
            buf += pidstr;
            j += length;
            free(pidstr);
            i++; // need to skip second $
          }
          else {
            // Get value of specified environment variable
            const char *envvar = getenv(expandbuf.c_str());
            if(envvar) {
              // getenv returned something, add it to the string
              int length = strlen(envvar);
              buf += envvar;
              j += length;
            }
          }
        }
        else if(filename[expandstart] == '~') {
          // Note that we always expect j==0 here because anything else is not allowed
          if(i >= filename.length() || filename[i] == '/') {
            // Home directory
            if(k==0) {
              // Current user's home directory
              const char *envvar = getenv("HOME");
              if(envvar) {
                buf += envvar;
                j = strlen(envvar);
              }
            }
            else {
              // Some other user's home directory
              passwd *userinfo;
              userinfo = getpwnam(expandbuf.c_str());
              if(userinfo) {
                buf += userinfo->pw_dir;
                j = strlen(userinfo->pw_dir);
              }
              else {
                j = i;
                buf.assign(filename, 0, j);
              }
            }
          }
          else {
            j = i;
            buf.assign(filename, 0, j);
          }
        }
        expanding = false;
      }
      else {
        expandbuf += filename[i];
        k++;
      }
    }

    if(i >= filename.length())
      break;

    if(!expanding) {
      if(filename[i] == '$' || (i==0 && filename[i] == '~')) {
        // Start of something to expand
        k = 0;
        expandstart = i;
        expanding = true;
      }
      else {
        buf += filename[i];
        j++;
      }
    }

    i++;
  }

  return buf;
}

// Convert a string value into a boolean
bool strtobool(const char *str) {
  if(strcasecmp(str, "true") == 0 || strcasecmp(str, "on") == 0 || strcasecmp(str, "yes") == 0 || atoi(str) != 0)
    return true;
  else
    return false;
}

// Count the number of instances of character c in string str
int countstr(const char *str, char c) {
  int count = 0;
  const char *p = str;
  while(true) {
    p = strchr(p, c);
    if(p) {
      count++;
      p++;
    }
    else
      break;
  }
  return count;
}

int statfile(const char *filename) {
  struct stat buf;

  if(stat(filename, &buf) != 0)
    return errno;
  else
    return 0;
}

bool isEmptyString(const char *str) {
  for(unsigned int i=0;i<strlen(str);i++) {
    if(!isspace(str[i]))
      return false;
  }
  return true;
}

int stylestrtoint(const string &style) {
  if(lowercase(style) == "tiled")
    return IMAGESTYLE_TILED;
  else
    return IMAGESTYLE_NORMAL;
}

string styleinttostr(int style) {
  if(style == IMAGESTYLE_TILED)
    return "tiled";
  else
    return "normal";
}

int autostrtoint(const string &autostr) {
  if(autostr == "on")
    return AUTOPROPERTY_ON;
  else if(autostr == "off")
    return AUTOPROPERTY_OFF;
  else
    return AUTOPROPERTY_AUTO;
}

string autointtostr(int autovalue) {
  if(autovalue == AUTOPROPERTY_ON)
    return "on";
  else if(autovalue == AUTOPROPERTY_OFF)
    return "off";
  else
    return "auto";
}

string lowercase(const string &str) {
  string lstr = str;
  transform (lstr.begin(), lstr.end(), lstr.begin(), (int(*)(int)) tolower);
  return lstr;
}

void delayms(unsigned int milliseconds) {
  struct timespec ts;
  ts.tv_sec = 0;
  ts.tv_nsec = milliseconds * 1000000; // 1/2 second
  nanosleep(&ts, NULL);
}

int strtoint(const std::string &str) {
  return atoi(str.c_str());
}

// Check if the specified file exists. Prints an error and returns false if it doesn't.
bool checkFileExists(const char *filename) {
  int rc = statfile(filename);

  switch(rc) {
    case ENOENT:
      cerr << "Could not find file '" << filename << "'" << endl;
      return false;
      break;
    case ENOTDIR:
      cerr << "Path in '" << filename << "' does not exist" << endl;
      return false;
      break;
    case EACCES:
      cerr << "Access denied to file '" << filename << "'" << endl;
      return false;
      break;
    case ENAMETOOLONG:
      cerr << "File name '" << filename << "' is too long" << endl;
      return false;
      break;
    default:
      return true;
  }
}

bool checkRegExp(const char *str, const char *regex_str) {
  int rc;
  regex_t *regex = NULL;

  /* Allocate space for the regular expression */
  regex = (regex_t *) malloc(sizeof(regex_t));
  memset(regex, 0, sizeof(regex_t));
  /* Compile regular expression */
  rc = regcomp(regex, regex_str, REG_EXTENDED);
  if(rc!=0) {
    /* An error occurred, get and print the error */
    size_t length;
    char *buffer;
    length = regerror(rc, regex, NULL, 0);
    buffer = (char *)malloc(length);
    regerror(rc, regex, buffer, length);
    fprintf(stderr, "%s\n", buffer);
    free(buffer);
    regfree(regex);
    free(regex);
    return false;
  }

  /* Execute regular expression */
  rc = regexec(regex, str, 0, NULL, REG_NOSUB);
  regfree(regex);
  free(regex);
  if(rc==0)
    return true;
  else
    return false;
}
