/***************************************************************************
                          widgetbase.cpp  -  base class for widgets
                             -------------------
    begin                : Sun Sep 21 2003
    copyright            : (C) 2003 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstring>
#include <cstdlib>
#include <iostream>

#include "widgetbase.h"
#include "utils.h"
#include "pagebase.h"
#include "memberfunctionproperty.h"
#include "membermethodhandler.h"
#include "../errors.h"
 
using namespace std;

extern const char *TYPESTR_PAGE;

WidgetBase::WidgetBase(void) {
  // Dummy constructor
}

WidgetBase::WidgetBase(int x, int y) {
  this->x = x;
  this->y = y;
  this->tabstop = false; // default
  this->focused = false;
  this->visible = true;
  this->origin = ORIGIN_DYNAMIC;

  MemberFunctionProperty<WidgetBase> *mp;
  mp = new MemberFunctionProperty<WidgetBase>("x", this, &WidgetBase::pget_X, &WidgetBase::pset_X, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<WidgetBase>("y", this, &WidgetBase::pget_Y, &WidgetBase::pset_Y, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<WidgetBase>("width", this, &WidgetBase::pget_Width, &WidgetBase::pset_Width, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<WidgetBase>("height", this, &WidgetBase::pget_Height, &WidgetBase::pset_Height, false);
  AddProperty(mp);
  mp = new MemberFunctionProperty<WidgetBase>("visible", this, &WidgetBase::pget_Visible, &WidgetBase::pset_Visible, false);
  AddProperty(mp);
  
  MemberMethodHandler<WidgetBase> *mh;
  mh = new MemberMethodHandler<WidgetBase>("setfocus", this, 0, &WidgetBase::m_SetFocus);
  AddMethod(mh);
}

WidgetBase::~WidgetBase(void) {
  // Nothing here yet
}

void WidgetBase::SetParent(WidgetBase *p) {
  this->parent = p;
}

bool WidgetBase::GetTabStop(void) {
  return tabstop;
}

void WidgetBase::SetTabStop(bool value) {
  tabstop = value;
}

int WidgetBase::GetX(void) {
  return x;
}

int WidgetBase::GetY(void) {
  return y;
}

void WidgetBase::SetX(int value) {
  this->SetPosition(value, GetY());
}

void WidgetBase::SetY(int value) {
  this->SetPosition(GetX(), value);
}

void WidgetBase::SetPosition(int x, int y) {
  this->x = x;
  this->y = y;
}

int WidgetBase::GetWidth(void) {
  return (int)w;
}

int WidgetBase::GetHeight(void) {
  return (int)h;
}

void WidgetBase::SetWidth(int value) {
  this->w = value;
}

void WidgetBase::SetHeight(int value) {
  this->h = value;
}

void WidgetBase::GotFocus(void) {
  focused = true;
}

void WidgetBase::LostFocus(void) {
  focused = false;
}

int WidgetBase::Clone(WidgetBase *newwidget) {
  // Copy property values
  // Note: override this in child objects to copy data that isn't available through properties
  int propcount = GetPropertyCount();
  for(int i=0; i<propcount; i++) {
    CustomProperty *prop = GetProperty(i);
    if(prop) {
      if(!(prop->IsReadOnly())) {
        int rc = newwidget->SetPropertyValue(prop->GetName(), prop->GetValue());
        if(rc != EBERR_SUCCESS) {
          cerr << "WidgetBase::Clone: failed to set property #" << i << " from object " << GetName() << " of type " << GetType() << endl;
          return rc;
        }  
      }
    }
    else {
      cerr << "WidgetBase::Clone: failed to retrieve property #" << i << " from object " << GetName() << " of type " << GetType() << endl;
      return -1;
    }
  }
  return EBERR_SUCCESS;
}

bool WidgetBase::StandardAction(int actiontype) {
  return false;
}

PageBase *WidgetBase::FindParentPage(void) {
  if(!parent)
    return NULL;
  else if(!strcmp(parent->GetType(), TYPESTR_PAGE))
    return dynamic_cast<PageBase *>(parent);
  else
    return parent->FindParentPage();
}

bool WidgetBase::GetVisible(void) {
  return visible;
}

void WidgetBase::SetOrigin(int origin) {
  this->origin = origin;
}

int WidgetBase::GetOrigin(void) {
  return origin;
}

// Property procedures

// -- x

Variant WidgetBase::pget_X(void) {
  return anytovariant(x);
}

int WidgetBase::pset_X(Variant value) {
  SetX(value);
  return EBERR_SUCCESS;
}

// -- y

Variant WidgetBase::pget_Y(void) {
  return anytovariant(y);
}

int WidgetBase::pset_Y(Variant value) {
  SetY(value);
  return EBERR_SUCCESS;
}

// -- width

Variant WidgetBase::pget_Width(void) {
  return anytovariant(w);
}

int WidgetBase::pset_Width(Variant value) {
  SetWidth(value);
  return EBERR_SUCCESS;
}

// -- height

Variant WidgetBase::pget_Height(void) {
  return anytovariant(h);
}

int WidgetBase::pset_Height(Variant value) {
  SetHeight(value);
  return EBERR_SUCCESS;
}

// -- visible

Variant WidgetBase::pget_Visible(void) {
  return anytovariant(visible);
}

int WidgetBase::pset_Visible(Variant value) {
  SetVisible(value);
  return EBERR_SUCCESS;
}

/* Method functions */

Variant WidgetBase::m_SetFocus(int numargs, Variant args[]) {
  PageBase *parentpage = FindParentPage();
  if(parentpage)
    parentpage->SetFocus( this );
  return VARNULL;
}
