/***************************************************************************
                          gui_extlabel.cpp  -  extended SDL_gui label class
                             -------------------
    begin                : Thu Aug 22 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstring>
#include <cstdlib>

#include "SDL_gui.h"
#include "../../basewidgets/utils.h"
#include "gui_extensions.h"

using namespace std;

GUI_ExtLabel::GUI_ExtLabel(const char *aname, int x, int y, int w, int h, GUI_Font *afont, const char *s)
: GUI_Label(aname, x, y, w, h, afont, s)
{
  textstring = strdup(s);
  wordwrap = false;
  autosize = false;
}

GUI_ExtLabel::~GUI_ExtLabel()
{
	free(textstring);
  // We need to do this because SDL_gui's label doesn't check before calling decref on text
  // (This happens if the label is empty)
  if(!text)
  	text = font->RenderFast("bob", textcolor);
}

void GUI_ExtLabel::Update(int force) {
  if (parent==0)
		return;
	
  if (force) {
		if (flags & WIDGET_TRANSPARENT)
			parent->Erase(&area);
		
    if (text) {
      SDL_Rect dr = {0,0,0,0};
			SDL_Rect sr = {0,0,0,0};
			SDL_Rect clip = {0,0,0,0};
			
			clip = parent->GetArea();
			clip.x = clip.y = sr.x = sr.y = 0;
      char *linedstr = strdup(textstring);
      char *bufptr = 0;
      GUI_Surface *line;
      int maxwidth = 0;

      /* Work out width of text */
      if (wordwrap && strchr(linedstr, ' ') != NULL) {
        bufptr = linedstr;  // start of string
        char *prevword = bufptr;
        char *linestart = bufptr;
        while(true) {
          bufptr = strchr(bufptr+1, ' ');

          if(bufptr)
            bufptr[0] = '\0'; // we're only interested in the line to this point
          // Render the line, so we can find out how wide it is
          if(linestart && strlen(linestart) > 0) {
            line = font->RenderFast(linestart, textcolor);
            if(bufptr)
              bufptr[0] = ' ';
            if(line->GetWidth() > area.w) {
              if(prevword == linedstr && bufptr != NULL) {
                // The first word is really long, so ignore it
                bufptr[0] = '\n';
                linestart = bufptr;
                if(line->GetWidth() > maxwidth)
                  maxwidth = line->GetWidth();
              }
              else {
                // Wrap at the start of the last word
                prevword[0] = '\n';
                linestart = prevword;
              }
            }
            else {
              if(line->GetWidth() > maxwidth)
                maxwidth = line->GetWidth();
            }
            line->DecRef();
          }
          prevword = bufptr;

          if(!bufptr)
            break;  // done
        }
        dr.w = maxwidth;
  			dr.h = text->GetHeight() * (countstr(linedstr, '\n') + 1);
      }
      else {
        int numlines = countstr(linedstr, '\n') + 1;
        if(numlines > 1) {
          // Manual-newline
          dr.h = text->GetHeight() * numlines;
          char *p = linedstr; // need to make a copy because strsep modifies the pointer
          while(true) {
            bufptr = strsep(&p, "\n");
            if(!bufptr)
              break;
            if(strlen(bufptr) > 0) {
              line = font->RenderFast(bufptr, textcolor);
              if(line->GetWidth() > maxwidth)
                maxwidth = line->GetWidth();
              line->DecRef();
            }
          }
          free(linedstr);
          linedstr = strdup(textstring); // reset
          dr.w = maxwidth;
        }
        else {
          // Single line
    			dr.h = text->GetHeight();
          dr.w = text->GetWidth();
        }
      }

      if(autosize) {
        if(!wordwrap)
          area.w = dr.w;
        area.h = dr.h;
      }

      sr.w = area.w;
      sr.h = area.h;

      // Set first position
 			switch (flags & WIDGET_HORIZ_MASK) {
 				case WIDGET_HORIZ_CENTER:
 				case WIDGET_HORIZ_STRETCH: /* FIXME: stretch not implemented */
 					dr.x = area.x + (area.w - dr.w) / 2;
 					break;
 				case WIDGET_HORIZ_LEFT:
 					dr.x = area.x;
 					break;
 				case WIDGET_HORIZ_RIGHT:
 					dr.x = area.x + area.w - dr.w;
 					break;
 			}
 			switch (flags & WIDGET_VERT_MASK) {
 				case WIDGET_VERT_CENTER:
 				case WIDGET_VERT_STRETCH: /* FIXME: stretch not implemented */
 					dr.y = area.y + (area.h - dr.h) / 2;
 					break;
 				case WIDGET_VERT_TOP:
 					dr.y = area.y;
 					break;
 				case WIDGET_VERT_BOTTOM:
 					dr.y = area.y + area.h - dr.h;
 					break;
 			}

      // Draw the text
      if (GUI_ClipRect(&sr, &dr, &clip)) {
				if(strchr(linedstr, '\n')) {
          // Multi-line
          char *p = linedstr; // need to make a copy because strsep modifies the pointer
          while(true) {
            bufptr = strsep(&p, "\n");
            if(!bufptr || dr.y > (area.y + area.h))
              break;
            if(strlen(bufptr) > 0) {
              // Render the font
              line = font->RenderQuality(bufptr, textcolor);
              // Figure out positioning
              dr.w = line->GetWidth();
              dr.h = line->GetHeight();
         			switch (flags & WIDGET_HORIZ_MASK) {
         				case WIDGET_HORIZ_CENTER:
         				case WIDGET_HORIZ_STRETCH: /* FIXME: stretch not implemented */
         					dr.x = area.x + (area.w - dr.w) / 2;
                  // Make adjustments for lines that are too wide
                  if(dr.x - area.x < 0) {
                    sr.x = area.x - dr.x;
                    dr.x += (area.x - dr.x);
                  }
                  else
                    sr.x = 0;
         					break;
         				case WIDGET_HORIZ_LEFT:
         					dr.x = area.x;
         					break;
         				case WIDGET_HORIZ_RIGHT:
         					dr.x = area.x + (area.w - dr.w);
                  // Make adjustments for lines that are too wide
                  if(dr.x - area.x < 0) {
                    sr.x = area.x - dr.x;
                    dr.x += (area.x - dr.x);
                  }
                  else
                    sr.x = 0;
         					break;
         			}
        			sr.h = area.h - (dr.y - area.y);
              // Draw the font to the screen
              DrawSurface(line, &sr, &dr);
              // Done with it, so release it
              line->DecRef();
            }
            else
              dr.h = text->GetHeight(); // FIXME: Hack!

            // Increment the vertical
            dr.y += dr.h;
          }
        }
        else {
          // Single line
          // FIXME: this code needs cleaning up
    			sr.h = area.h - (dr.x - area.x);
          DrawClipped(text, parent, area, dr.x, dr.y);
//          parent->Draw(text, &sr, &dr);
        }
      }
      free(linedstr);
		}
	}
}

void GUI_ExtLabel::DrawSurface(GUI_Surface *surf, SDL_Rect *sr, SDL_Rect *dr) {
  parent->Draw(surf, sr, dr);
}

void GUI_ExtLabel::SetFont(GUI_Font *afont)
{
	if (GUI_ObjectKeep((GUI_Object **) &font, afont)) {
  	if (text) text->DecRef();
	  text = font->RenderQuality(textstring, textcolor);
 		MarkChanged();
  }
}

void GUI_ExtLabel::SetTextColor(int r, int g, int b)
{
	textcolor.r = r;
	textcolor.g = g;
	textcolor.b = b;
	if(text)
    text->DecRef();
	text = font->RenderQuality(textstring, textcolor);
}

void GUI_ExtLabel::SetText(const char *s)
{
	if(text)
    text->DecRef();
	text = font->RenderQuality(s, textcolor);
  free(textstring);
  textstring = strdup(s);
	MarkChanged();
}

void GUI_ExtLabel::SetAutoSize(bool autosize) {
  this->autosize = autosize;
}

void GUI_ExtLabel::SetWordWrap(bool wordwrap) {
  this->wordwrap = wordwrap;
}
