/***************************************************************************
          gui_extscrollbar.cpp  - extended SDL_gui scrollbar widget
                             -------------------
    begin                : Sun Nov 10 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <assert.h>
#include <cstring>
#include <cstdlib>

#include "SDL_gui.h"
#include "gui_extensions.h"
#include "../../basewidgets/utils.h"

using namespace std;

GUI_ExtScrollBar::GUI_ExtScrollBar(const char *aname, int x, int y, int w, int h)
: GUI_ScrollBar(aname, x, y, w, h)
{
  maximum = 100;
  minimum = 0;
  value = 0;
  page_step = 10;
  background_style = IMAGESTYLE_NORMAL;
	knob->Fill(NULL, 0xFFFFFFFF);
  knob_focus = NULL;
  reversed = false;
}

GUI_ExtScrollBar::~GUI_ExtScrollBar(void)
{
  if(knob_focus)
    knob_focus->DecRef();
}

void GUI_ExtScrollBar::SetPosition(int x, int y)
{
  area.x = x;
  area.y = y;
  MarkChanged();
}

void GUI_ExtScrollBar::SetWidth(int w)
{
  area.w = w;
  MarkChanged();
}

void GUI_ExtScrollBar::SetHeight(int h)
{
  area.h = h;
  MarkChanged();
}

void GUI_ExtScrollBar::SetValue(int value)
{
  this->value = value;
  RecalcPosition();
  MarkChanged();
}

int GUI_ExtScrollBar::GetValue(void)
{
  return value;
}

void GUI_ExtScrollBar::SetPageStep(int step)
{
  if(step > 0)
  	this->page_step = step;
}

void GUI_ExtScrollBar::SetReversed(bool reversed)
{
  this->reversed = reversed;
  RecalcPosition();
  MarkChanged();
}

int GUI_ExtScrollBar::GetMaximum(void)
{
  return maximum;
}

void GUI_ExtScrollBar::SetMaximum(int maximum)
{
  this->maximum = maximum;
  RecalcPosition();
  MarkChanged();
}

int GUI_ExtScrollBar::GetMinimum(void)
{
  return minimum;
}

void GUI_ExtScrollBar::SetMinimum(int minimum)
{
  this->minimum = minimum;
  RecalcPosition();
  MarkChanged();
}

void GUI_ExtScrollBar::RecalcPosition(void)
{
  int max_visual;

  if (area.h > area.w)
  {
    max_visual = area.h;
    if(knob)
      max_visual -= knob->GetHeight();
  }
  else
  {
    max_visual = area.w;
    if(knob)
      max_visual -= knob->GetWidth();
  }

  if(reversed)
  	position = (int)((((float)(maximum - minimum) - (value - minimum)) / (maximum - minimum)) * max_visual);
  else
  	position = (int)(((float)(value - minimum) / (maximum - minimum)) * max_visual);
}

void GUI_ExtScrollBar::RecalcValue(void)
{
  int max_visual;

  if (area.h > area.w)
  {
    max_visual = area.h;
    if(knob)
      max_visual -= knob->GetHeight();
  }
  else
  {
    max_visual = area.w;
    if(knob)
      max_visual -= knob->GetWidth();
  }
	
  value = (int)(((float)position/max_visual) * (maximum - minimum)) + minimum;
  if(reversed)
    value = (maximum - value) + minimum;
}

void GUI_ExtScrollBar::Update(int force)
{
  GUI_Surface *surface = NULL;

	if (parent == 0)
		return;
	
	if (force)
	{
		if (flags & WIDGET_TRANSPARENT)
			parent->Erase(&area);
		
    if(background) {
      if(background_style == IMAGESTYLE_NORMAL)
        DrawClipped(background, parent, area, area.x, area.y);
      else if(background_style == IMAGESTYLE_TILED)
        parent->TileImage(background, &area, 0, 0);
    }

		if ((flags & WIDGET_HAS_FOCUS) && knob_focus)
      surface = knob_focus;
    else
      surface = knob;

		if (surface)
    {
      if (area.h > area.w)
        DrawClipped(surface, parent, area, area.x, area.y + position);
      else
        DrawClipped(surface, parent, area, area.x + position, area.y);
    }
	}
}

int GUI_ExtScrollBar::Event(const SDL_Event *event, int xoffset, int yoffset)
{
	switch (event->type)
	{
		case SDL_MOUSEMOTION:
		{
			int x = event->button.x - area.x;
			int y = event->motion.y - area.y;
			if (tracking_on)
			{
        int positionmax;
				
        if (area.h > area.w)
        {
	  			position = tracking_pos + y - tracking_start;
          positionmax = area.h - knob->GetHeight();
        }
        else
        {
	  			position = tracking_pos + x - tracking_start;
          positionmax = area.w - knob->GetWidth();
        }

				if (position < 0)
					position = 0;
				if (position > positionmax)
					position = positionmax;
        RecalcValue();
				MarkChanged();
				if (moved_callback)
					moved_callback->Call(this);
				return 1;
			}
			break;
		}
		case SDL_MOUSEBUTTONDOWN:
		{
			int x = event->button.x - area.x;
      int y = event->button.y - area.y;
			if (flags & WIDGET_INSIDE && x > 0 && x < area.w) // this is my fault for mucking about with WIDGET_INSIDE
			{
				if(event->button.button == SDL_BUTTON_LEFT) {

  				/* if the cursor is inside the knob, start tracking */
  				if (area.h > area.w) {
            if (y >= position && y < position + knob->GetHeight())
    				{
    					tracking_on = 1;
    					tracking_start = y;
    					tracking_pos = position;
    				}
          }
          else
          {
            if (x >= position && x < position + knob->GetWidth())
    				{
    					tracking_on = 1;
    					tracking_start = x;
    					tracking_pos = position;
    				}
          }
  				return 1;
        }
        else if(event->button.button == 4) {
          // Scroll wheel up
					value -= 3;
					if (value < minimum)
						value = minimum;
          RecalcPosition();
  				if (moved_callback)
  					moved_callback->Call(this);
  				MarkChanged();
  				return 1;
        }
        else if(event->button.button == 5) {
          // Scroll wheel down
					value += 3;
					if (value > maximum)
						value = maximum;
          RecalcPosition();
  				if (moved_callback)
  					moved_callback->Call(this);
  				MarkChanged();
  				return 1;
        }
			}
			break;
		}
		case SDL_MOUSEBUTTONUP:
		{
			int y = event->button.y - area.y;
			int x = event->button.x - area.x;
 			if (tracking_on)
 			{
        tracking_on = 0;
 			}
 			else if((flags & WIDGET_INSIDE) && pointinrect(event->button.x, event->button.y, area) && event->button.button == SDL_BUTTON_LEFT)
 			{
        bool above_knob = false;
        bool below_knob = false;

        if ((area.h > area.w && y < position) || (area.h < area.w && x < position))
          above_knob = true;
        else if ((area.h > area.w && y >= position + knob->GetHeight()) || (area.h < area.w && x >= position + knob->GetWidth()))
          below_knob = true;

        if((above_knob && !reversed) || (below_knob && reversed))
        {
          value -= page_step;
          if (value < minimum)
            value = minimum;
        }
        else if((below_knob && !reversed) || (above_knob && reversed))
        {
          value += page_step;
          if (value > maximum)
            value = maximum;
        }

        RecalcPosition();

        if (moved_callback)
          moved_callback->Call(this);
          
        MarkChanged();
        return 1;
      }
      break;
    }
  }
  return GUI_Drawable::Event(event, xoffset, yoffset);
}

void GUI_ExtScrollBar::SetBackgroundStyle(int style)
{
  if(style == IMAGESTYLE_NORMAL || style == IMAGESTYLE_TILED) {
    background_style = style;
    MarkChanged();
  }
}

int GUI_ExtScrollBar::GetBackgroundStyle(void)
{
  return background_style;
}

void GUI_ExtScrollBar::SetKnobFocusImage(GUI_Surface *image)
{
	if (GUI_ObjectKeep((GUI_Object **) &knob_focus, image))
		MarkChanged();
}

