/***************************************************************************
                     gui_listbox.cpp - SDL_gui listbox widget
                             -------------------
    begin                : Thu Nov 7 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <assert.h>
#include <cstring>
#include <cstdlib>

#include "SDL.h"
#include "SDL_gui.h"

#include "gui_extensions.h"
#include "../../sdluserevents.h"
#include "../../basewidgets/utils.h"

using namespace std;

GUI_ListBox::GUI_ListBox(const char *aname, int x, int y, int w, int h, GUI_Font *afont, int size)
: GUI_Widget(aname, x, y, w, h), font(afont)
{
	SetTransparent(1);
	
	textcolor.r = 255;
	textcolor.g = 255;
	textcolor.b = 255;
	textcolor.unused = 255;

	selitemtextcolor.r = 255;
	selitemtextcolor.g = 255;
	selitemtextcolor.b = 0;
	selitemtextcolor.unused = 255;

  selected_item_image = NULL;
  selected_item_focus_image = NULL;
  selected_item_style = IMAGESTYLE_NORMAL;
  background_image = NULL;
  background_focus_image = NULL;
  background_style = IMAGESTYLE_NORMAL;

	font->IncRef();

  topindex = 0;
  selectedindex = -1;
  tracking = false;
  scrolling = false;
  visibleitems = 0;
  borderx = 0;
  bordery = 0;
  itemborderx = 0;
  itembordery = 0;
  useritemheight = -1; // automatic (text height)
  scrollbarmode = AUTOPROPERTY_AUTO;
  scrollbarvisible = false;

  scrollbar = new GUI_ExtScrollBar("scroll", x+w-16-borderx, y+bordery, 16, h-bordery*2);
  RecalcItemHeight();

//  scrollbar->ClearFlags(WIDGET_TRANSPARENT);
  scrollbar->SetValue(0);
	
  GUI_EventHandler<GUI_ListBox> *cb = new GUI_EventHandler<GUI_ListBox>(this, &GUI_ListBox::OnScrollBarMove);
  scrollbar->SetMovedCallback(cb);
	cb->DecRef();

	choose_callback = NULL;
	selection_callback = NULL;
  lastclicktime = 0;
}

GUI_ListBox::~GUI_ListBox()
{
	font->DecRef();
  if(background_image)
  	background_image->DecRef();
  if(background_focus_image)
  	background_focus_image->DecRef();
  if(selected_item_image)
  	selected_item_image->DecRef();
	if (selection_callback) selection_callback->DecRef();
	if (choose_callback) choose_callback->DecRef();
  while(listitems.size() > 0) {
    char *str = listitems.back();
    listitems.pop_back();
    free(str);
  }
}

void GUI_ListBox::Update(int force)
{
	if (parent==0)
		return;
	
	if (force)
	{
		GUI_Surface *surface = NULL;
		
		if (flags & WIDGET_TRANSPARENT)
			parent->Erase(&area);
		
		// draw the background
		if((flags & WIDGET_HAS_FOCUS) && (background_focus_image != NULL))
		{
			surface = background_focus_image;
		}
		else
		{
  		surface = background_image;
		}

    if(surface) {
      if(background_style == IMAGESTYLE_NORMAL)
        DrawClipped(surface, parent, area, area.x, area.y);
      else if(background_style == IMAGESTYLE_TILED)
        parent->TileImage(surface, &area, 0, 0);
    }
		
    for(unsigned int i=0; i+topindex<listitems.size() && i < visibleitems+1; i++) {
      int itemx = area.x + borderx;
      int itemy = area.y + bordery + (i * itemheight);

      SDL_Rect clip = {itemx, itemy, area.w - (borderx * 2), itemheight};
      if(clip.y + clip.h > area.y + area.h - bordery)   // if the last item is smaller than the item height
        clip.h = area.h - (itemy - area.y) - bordery;

      if(scrollbarvisible)
        clip.w = clip.w - scrollbar->GetWidth();

      if(i+topindex == selectedindex) {
        // The selected item
        if((flags & WIDGET_HAS_FOCUS) && (selected_item_focus_image != NULL))
          surface = selected_item_focus_image;
        else
          surface = selected_item_image;

        if(surface) {
          if(selected_item_style == IMAGESTYLE_NORMAL)
            DrawClipped(surface, parent, clip, clip.x, clip.y);
          else if(selected_item_style == IMAGESTYLE_TILED)
            parent->TileImage(surface, &clip, 0, 0);
        }
        
        // Now render the text, as selected
    		surface = font->RenderQuality(listitems[i+topindex], selitemtextcolor);
      }
      else {
        // Now render the text, as normal
    		surface = font->RenderQuality(listitems[i+topindex], textcolor);
      }
      
  		if (surface != NULL)
  		{
        // Ensure text is centred within available item height
        int textx = itemx + itemborderx;
        int texty = itemy + ((itemheight - surface->GetHeight()) / 2);
        // Draw the text
        clip.x = clip.x + itemborderx;
        clip.y = clip.y + itembordery;
        clip.w = clip.w - (itemborderx * 2);
        clip.h = clip.h - (itembordery * 2);
        if(clip.y + clip.h > area.y + area.h)   // if the last item is smaller than the item height
          clip.h = area.h - (clip.y - area.y);
        DrawClipped(surface, parent, clip, textx, texty);
        surface->DecRef();
  		}
    }
	}

  if(scrollbarvisible) {
    if(scrollbar->GetParent() == NULL && GetParent() != NULL)
      ((GUI_Panel *)GetParent())->AddWidget(scrollbar);
    else
      scrollbar->Update(force);
  }
  else if(scrollbar->GetParent() != NULL)
    ((GUI_Panel *)GetParent())->RemoveWidget(scrollbar);
}

int GUI_ListBox::ItemAtPoint(int x, int y) {
  if(y >= area.y + bordery && y < area.y + area.h - bordery)
    return ((y-area.y-bordery) / itemheight) + topindex;
  else
    return -1;
}

int GUI_ListBox::Event(const SDL_Event *event, int xoffset, int yoffset)
{
  SDL_Rect a = scrollbar->GetArea();
  if(event->type == SDL_MOUSEBUTTONDOWN || event->type == SDL_MOUSEBUTTONUP) {
    if(scrollbarvisible && pointinrect(event->button.x, event->button.y, a))
      // This event belongs to the scrollbar - but just let it go.
      // If we try to pass it to the scrollbar then it gets the event twice.
      // On the other hand we don't want it, so we return 0 indicating it hasn't
      // been processed yet.
      return 0;
    else {
      // Mouse button on listbox
      if(event->type == SDL_MOUSEBUTTONDOWN && (flags & WIDGET_INSIDE)) {
        if(event->button.button == SDL_BUTTON_LEFT || event->button.button == SDL_BUTTON_RIGHT) {
          tracking = true;
          int index = ItemAtPoint(event->button.x,event->button.y);
          if(index >= 0 && index < listitems.size()) {
            int lastselectedindex = selectedindex;
            SetSelectedIndex(index);

            Uint32 ticks = SDL_GetTicks();
            if(lastselectedindex == selectedindex && lastclicktime > 0 && ticks - lastclicktime < DOUBLECLICK_TIME
                && choose_callback != NULL) {
              choose_callback->Call(this);
              lastclicktime = 0;
            }
            else
              lastclicktime = ticks;
          }
        }
        else if(event->button.button == 4)
          SetTopIndex(topindex-3);
        else if(event->button.button == 5)
          SetTopIndex(topindex+3);
      }
      else if(event->type == SDL_MOUSEBUTTONUP)
        tracking = false;
    }
  }
  else if(event->type == SDL_MOUSEMOTION) {
    if(scrollbarvisible && pointinrect(event->motion.x, event->motion.y, a)) {
      // don't return here, so that listbox gets focus if mouse moves over scrollbar
    }
    else if(flags & WIDGET_INSIDE) {
      if(tracking) {
        int index = ItemAtPoint(event->button.x,event->button.y);
        if(index >= 0)
          SetSelectedIndex(index);
      }
    }
  }

/*  if(flags & WIDGET_HAS_FOCUS) {
	  if (event->type == SDL_KEYDOWN)
  	{
  		int key = event->key.keysym.sym;
  		int ch = event->key.keysym.unicode;

  		if (key == SDLK_UP)
      {
        SelectPrevious(1);
        return 1;
      }
  		if (key == SDLK_DOWN)
  		{
        SelectNext(1);
        return 1;
  		}
  		if (key == SDLK_PAGEUP)
  		{
        SelectPrevious(visibleitems);
        return 1;
  		}
  		if (key == SDLK_PAGEUP)
  		{
        SelectNext(visibleitems);
        return 1;
  		}
  		if (key == SDLK_HOME)
  		{
        SelectFirst();
        return 1;
  		}
  		if (key == SDLK_END)
  		{
        SelectLast();
        return 1;
  		}
  		if (ch >= 32 && ch <= 126)
  		{
        // ASCII
  		}
  	}
  }*/

	return GUI_Widget::Event(event, xoffset, yoffset);
}

void GUI_ListBox::SetFont(GUI_Font *afont)
{
	GUI_ObjectKeep((GUI_Object **) &font, afont);
  RecalcItemHeight();
  MarkChanged();
}

void GUI_ListBox::SetTextColor(int r, int g, int b)
{
	textcolor.r = r;
	textcolor.g = g;
	textcolor.b = b;
  MarkChanged();
}

void GUI_ListBox::SetSelectedItemTextColor(int r, int g, int b)
{
	selitemtextcolor.r = r;
	selitemtextcolor.g = g;
	selitemtextcolor.b = b;
  MarkChanged();
}

void GUI_ListBox::SetScrollBarMode(int scrollbarmode)
{
  this->scrollbarmode = scrollbarmode;
  CheckScrollBarVisible();
}

void GUI_ListBox::SetScrollBarWidth(int width)
{
  scrollbar->SetPosition(area.x + area.w - width - borderx, area.y + bordery);
  scrollbar->SetWidth(width);
  MarkChanged();
}

void GUI_ListBox::SetBorderX(int pixels)
{
  this->borderx = pixels;
  RecalcItemHeight();
  scrollbar->SetPosition(area.x + area.w - scrollbar->GetWidth() - borderx, area.y + bordery);
  MarkChanged();
}

void GUI_ListBox::SetBorderY(int pixels)
{
  this->bordery = pixels;
  RecalcItemHeight();
  scrollbar->SetHeight(area.h - (bordery*2));
  scrollbar->SetPosition(area.x + area.w - scrollbar->GetWidth() - borderx, area.y + bordery);
  MarkChanged();
}

void GUI_ListBox::SetItemBorderX(int pixels)
{
  this->itemborderx = pixels;
  MarkChanged();
}

void GUI_ListBox::SetItemBorderY(int pixels)
{
  this->itembordery = pixels;
  MarkChanged();
}

void GUI_ListBox::SetItemHeight(int pixels)
{
  this->useritemheight = pixels;
  RecalcItemHeight();
  MarkChanged();
}

void GUI_ListBox::RecalcItemHeight(void) {
  // Get list item height
  if(useritemheight == -1) {
    SDL_Rect textsize = font->GetTextSize("testing, 123");
    itemheight = textsize.h;
  }
  else
    itemheight = useritemheight;

  visibleitems = (area.h - (bordery * 2)) / itemheight;
  SetupScrollBar();
}

void GUI_ListBox::AddItem(const char *text)
{
	int numitems = listitems.size();
  assert(text != NULL);
  listitems.push_back(strdup(text));
  SetupScrollBar();
  if(numitems==0)
    SetSelectedIndex(0);
  MarkChanged();
}

void GUI_ListBox::RemoveItem(int index)
{
  if(index >= 0 && index < listitems.size()) {
    vector<char *>::iterator iter = listitems.begin() + index;
    listitems.erase(iter);
    SetupScrollBar();
    // Check if we need to move the selection
    if(listitems.size() > 0) {
      if(selectedindex > listitems.size()-1)
        SelectLast();                         // Removed item was the last item
      else if(index < selectedindex)
        SelectPrevious(1);                    // Removed item was before the current item
      else if(index == selectedindex)
        InternalSetSelectedIndex(index, true); // Removed item was the current item. We don't change the index but we need to fire an event.
    }
    else
      selectedindex = -1;
    MarkChanged();
  }
}

void GUI_ListBox::Clear(void) {
  while(listitems.size() > 0) {
    char *str = listitems.back();
    listitems.pop_back();
    free(str);
  }
  topindex = 0;
  selectedindex = -1;
  SetupScrollBar();
  MarkChanged();
}

void GUI_ListBox::SetItemText(int index, const char *text) {
  if(index >= 0 && index < listitems.size()) {
    char *olditem = listitems[index];
    listitems[index] = strdup(text);
    free(olditem);
    if(index >= topindex && index < topindex + visibleitems)
      MarkChanged();
  }
}

void GUI_ListBox::SetBackgroundImage(GUI_Surface *surface)
{
	if (GUI_ObjectKeep((GUI_Object **) &background_image, surface))
		MarkChanged();
}

void GUI_ListBox::SetBackgroundFocusImage(GUI_Surface *surface)
{
	if (GUI_ObjectKeep((GUI_Object **) &background_focus_image, surface))
		MarkChanged();
}

void GUI_ListBox::SetBackgroundStyle(int style)
{
  if(style == IMAGESTYLE_NORMAL || style == IMAGESTYLE_TILED) {
    background_style = style;
    MarkChanged();
  }
}

void GUI_ListBox::SetSelectedItemImage(GUI_Surface *surface)
{
	if (GUI_ObjectKeep((GUI_Object **) &selected_item_image, surface))
		MarkChanged();
}

void GUI_ListBox::SetSelectedItemFocusImage(GUI_Surface *surface)
{
	if (GUI_ObjectKeep((GUI_Object **) &selected_item_focus_image, surface))
		MarkChanged();
}

void GUI_ListBox::SetSelectedItemStyle(int style)
{
  if(style == IMAGESTYLE_NORMAL || style == IMAGESTYLE_TILED) {
    selected_item_style = style;
    MarkChanged();
  }
}

void GUI_ListBox::SetScrollBarKnobImage(GUI_Surface *surface)
{
  scrollbar->SetKnobImage(surface);
}

void GUI_ListBox::SetScrollBarBackgroundImage(GUI_Surface *surface)
{
  scrollbar->SetBackgroundImage(surface);
}

void GUI_ListBox::SetScrollBarBackgroundStyle(int style)
{
  scrollbar->SetBackgroundStyle(style);
}

void GUI_ListBox::SetSelectedIndex(int index) {
  InternalSetSelectedIndex(index, false);
}

void GUI_ListBox::InternalSetSelectedIndex(int index, bool forceevent) {
  int lastindex = selectedindex;

  if(scrolling)
    return;
  scrolling = true;
  if(index >= 0 && index < listitems.size()) {
    selectedindex = index;
    if(selectedindex > topindex + (visibleitems - 1))
      topindex = selectedindex - (visibleitems - 1);
    else if(selectedindex < topindex)
      topindex = selectedindex;
    scrollbar->SetValue(topindex);
    MarkChanged();
		if (selection_callback != NULL && (selectedindex != lastindex || forceevent))
			selection_callback->Call(this);
  }
  scrolling = false;
}

int GUI_ListBox::GetSelectedIndex(void) {
  if(listitems.size() > 0)
    return selectedindex;
  else
    return -1;
}

int GUI_ListBox::GetVisibleItems(void) {
  return visibleitems;
}

void GUI_ListBox::SetTopIndex(int index) {
  if(listitems.size() > visibleitems) {
    if(index < 0)
      topindex = 0;
    else if(index > listitems.size() - visibleitems)
      topindex = listitems.size() - visibleitems;
    else
      topindex = index;
    scrollbar->SetValue(topindex);
    MarkChanged();
  }
}

void GUI_ListBox::SelectNext(int count) {
  if(selectedindex < listitems.size() - 1 && listitems.size() > 0) {
    // Select the next item
    if(selectedindex + count > listitems.size() - 1)
      SetSelectedIndex(listitems.size() - 1);
    else
      SetSelectedIndex(selectedindex + count);
  }
}

void GUI_ListBox::SelectPrevious(int count) {
  if(selectedindex > 0) {
    // Select the previous item
    if(selectedindex - count < 0)
      SetSelectedIndex(0);
    else
      SetSelectedIndex(selectedindex - count);
  }
}

void GUI_ListBox::SelectFirst(void) {
  if(selectedindex != 0)
    SetSelectedIndex(0);
}

void GUI_ListBox::SelectLast(void) {
  if(selectedindex != listitems.size() - 1)
    SetSelectedIndex(listitems.size() - 1);
}

void GUI_ListBox::OnScrollBarMove(GUI_Object *sender) {
  SetTopIndex(scrollbar->GetValue());
}

void GUI_ListBox::SetSelectionChangeCallback(GUI_Callback *callback)
{
	GUI_ObjectKeep((GUI_Object **) &selection_callback, callback);
}

void GUI_ListBox::SetChooseCallback(GUI_Callback *callback)
{
	GUI_ObjectKeep((GUI_Object **) &choose_callback, callback);
}

void GUI_ListBox::SetupScrollBar(void)
{
  if(listitems.size()>visibleitems)
    scrollbar->SetMaximum(listitems.size()-visibleitems);
  else
  {
    scrollbar->SetMaximum(0);
    scrollbar->SetValue(0);
  }
  scrollbar->SetPageStep(visibleitems);
  CheckScrollBarVisible();
}

void GUI_ListBox::CheckScrollBarVisible(void) {
  if(scrollbarmode == AUTOPROPERTY_OFF || (scrollbarmode == AUTOPROPERTY_AUTO && listitems.size() <= visibleitems))
    scrollbarvisible = false;
  else
    scrollbarvisible = true;
  MarkChanged();
}
