/***************************************************************************
                gui_textfield.cpp - SDL_gui textfield widget
              (based on TextEntry.cc from the SDL_gui project)
                             -------------------
    begin                : Sun Dec 8 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <assert.h>
#include <cstring>
#include <cstdlib>

#include "SDL_gui.h"
#include "gui_extensions.h"
#include "../../basewidgets/utils.h"
#include "../../basewidgets/textfieldwidgetbase.h"

GUI_TextField::GUI_TextField(const char *aname, int x, int y, int w, int h, GUI_Font *afont, int maxlength)
: GUI_Widget(aname, x, y, w, h), font(afont)
{
	SetTransparent(1);
	
  background_image = NULL;
  background_focus_image = NULL;
  background_style = IMAGESTYLE_NORMAL;

	textcolor.r = 255;
	textcolor.g = 255;
	textcolor.b = 255;
	textcolor.unused = 255;

  borderx = 0;
  bordery = 0;

	font->IncRef();
	
  cursorpos = 0;
  startoffset = 0;

	buffer_size = maxlength;
	buffer_index = 0;
	buffer = new char[maxlength+1];
	strcpy(buffer, "");
  validchars = strdup("");

	changed_callback = 0;
}

GUI_TextField::~GUI_TextField()
{
	font->DecRef();
  if(background_image)
  	background_image->DecRef();
  if(background_focus_image)
  	background_focus_image->DecRef();
	if(changed_callback)
    changed_callback->DecRef();
  free(validchars);
	delete [] buffer;
}

void GUI_TextField::Update(int force)
{
	if (parent==0)
		return;
	
	if (force)
	{
		GUI_Surface *surface;
		
		if (flags & WIDGET_TRANSPARENT)
			parent->Erase(&area);
		
		// draw the background
		if((flags & WIDGET_HAS_FOCUS) && (background_focus_image != NULL))
		{
			surface = background_focus_image;
		}
		else
		{
  		surface = background_image;
		}

    if(surface) {
      if(background_style == IMAGESTYLE_NORMAL)
        DrawClipped(surface, parent, area, area.x, area.y);
      else if(background_style == IMAGESTYLE_TILED)
        parent->TileImage(surface, &area, 0, 0);
    }
		
  	SDL_Rect clip = area;
    clip.x = clip.x + borderx;
    clip.y = clip.y + bordery;
    clip.w = clip.w - (borderx * 2);
    clip.h = clip.h - (bordery * 2);

		// Draw text
    surface = font->RenderQuality(buffer, textcolor);
		if (surface != NULL)
		{
      DrawClipped(surface, parent, clip, area.x + borderx - startoffset, area.y + ((area.h - surface->GetHeight()) / 2));
			surface->DecRef();
		}

    // Draw cursor
		if(flags & WIDGET_HAS_FOCUS) {
      char *cursortext;
      int cursorpx = 0;
      if(cursorpos > 0) {
        cursortext = strdup(buffer);
        cursortext[cursorpos] = '\0';
      }
      else
        cursortext = strdup("test");

      SDL_Rect textsize = font->GetTextSize(cursortext);

      if(cursorpos > 0)
        cursorpx = area.x + borderx + textsize.w - startoffset;
      else
        cursorpx = area.x + borderx;

    	surface = new GUI_Surface("cursor", SDL_SWSURFACE, 2, textsize.h + 2, 32, 0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000);
      surface->Fill(NULL, colortouint(textcolor));
      DrawClipped(surface, parent, clip, cursorpx, area.y + ((area.h - surface->GetHeight()) / 2));
			surface->DecRef();
      free(cursortext);
    }
	}
}

void GUI_TextField::Clicked(int x, int y)
{
	GUI_Screen *screen = GUI_GetScreen();
	
	if ((flags & WIDGET_HAS_FOCUS) == 0)
	{
		screen->SetFocusWidget(this);
	}
	MarkChanged();
}

int GUI_TextField::Event(const SDL_Event *event, int xoffset, int yoffset)
{
	if (event->type == SDL_KEYDOWN && flags & WIDGET_HAS_FOCUS)
	{
    // Note: this code is not actually used - key events never get as far as here
		int key = event->key.keysym.sym;
		int ch = event->key.keysym.unicode;
		
		if (key == SDLK_RIGHT)
		{
      if(cursorpos <= strlen(buffer)) {
        SetCursorPos(cursorpos+1);
      }
			return 1;
		}
		if (key == SDLK_LEFT)
		{
      if(cursorpos > 0) {
        SetCursorPos(cursorpos-1);
      }
			return 1;
		}
		if (key == SDLK_DELETE)
		{
      DeleteCurrChar();
			return 1;
		}
		if (key == SDLK_BACKSPACE)
		{
      Backspace();
			return 1;
		}
		if (key == SDLK_RETURN)
		{
/*			GUI_Screen *screen = GUI_GetScreen();
			screen->ClearFocusWidget();
			if (unfocus_callback)
				unfocus_callback->Call(this);	
			return 1;*/
		}
		if (ch >= 32 && ch <= 126)
		{
      SendChar(ch, INSMODE_INSERT);
			return 1;
		}
	}
	else if (event->type == SDL_MOUSEBUTTONDOWN && pointinrect(event->button.x, event->button.y, area)) {
    int cursorx = event->button.x - area.x - borderx;
    int lastwidth = 0;
    SDL_Rect textsize;
    char *cursortext = strdup(buffer);

    // Grab full width of text
    textsize = font->GetTextSize(cursortext);
    lastwidth = textsize.w;

    // Now iterate through and find the nearest character position to the point clicked
    for(int i=buffer_index-1; i>=0; i--) {
      cursortext[i] = '\0';
      textsize = font->GetTextSize(cursortext);
      if(textsize.w < cursorx) {
        // Get nearest position to that clicked
        if(cursorx - textsize.w < lastwidth - cursorx)
          cursorpos = i;
        else
          cursorpos = i+1;
        break;
      }
      lastwidth = textsize.w;
    }
    free(cursortext);
    MarkChanged();
    return 1;
  }
	
	return GUI_Widget::Event(event, xoffset, yoffset);
}

bool GUI_TextField::SendChar(char c, TextFieldInsertMode insmode)
{
	if (buffer_index < buffer_size)
	{
    if(strlen(validchars)) {
      // Do a regular expression check to see if the character should be allowed
      char *charstr;
      bool is_valid;
      charstr = (char *)malloc(2);
      snprintf(charstr, 2, "%c", c);
      is_valid = checkRegExp(charstr, validchars);

      free(charstr);
      if(!is_valid)
        return false; // not accepted
    }
    if(cursorpos==buffer_index) {
      if(insmode == INSMODE_BACKOVER && buffer_index > 0)
        buffer[buffer_index-1] = c;
      else
        buffer[buffer_index++] = c;
    }
    else {
      // Insert character
      char *bkup = strdup(buffer);
      if(insmode == INSMODE_INSERT)
        strncpy(buffer+cursorpos+1, bkup+cursorpos, buffer_index - cursorpos);
      buffer[cursorpos] = c;
      buffer_index++;
      free(bkup);
    }
		buffer[buffer_index] = 0;
    SetCursorPos(cursorpos+1);
    if(changed_callback)
      changed_callback->Call(this);
		MarkChanged();
	}
  return true;
}

void GUI_TextField::Backspace(void) {
 	if (buffer_index > 0 && cursorpos > 0)
 	{
    if(cursorpos<buffer_index) {
      char *bkup = strdup(buffer);
      strncpy(buffer+cursorpos-1, bkup+cursorpos, buffer_index - cursorpos);
      free(bkup);
    }
 		buffer[--buffer_index] = '\0';
    SetCursorPos(cursorpos-1);
    if(changed_callback)
      changed_callback->Call(this);
 		MarkChanged();
 	}
}

void GUI_TextField::DeleteCurrChar(void) {
  if(cursorpos < buffer_index) {
    char *bkup = strdup(buffer);
    strncpy(buffer+cursorpos, bkup+cursorpos+1, buffer_index - cursorpos);
		buffer[--buffer_index] = '\0';
    free(bkup);
    if(changed_callback)
      changed_callback->Call(this);
    MarkChanged();
  }
}

void GUI_TextField::SetFont(GUI_Font *afont)
{
	GUI_ObjectKeep((GUI_Object **) &font, afont);
  MarkChanged();
}

void GUI_TextField::SetTextColor(int r, int g, int b)
{
	textcolor.r = r;
	textcolor.g = g;
	textcolor.b = b;
  MarkChanged();
}

void GUI_TextField::SetText(const char *text)
{
	assert(text != NULL);

	if (strlen(text) < buffer_size)
	{
		strcpy(buffer, text);
		buffer_index = strlen(text);
	}
	MarkChanged();
}

const char *GUI_TextField::GetText(void)
{
	return buffer;
}

int GUI_TextField::GetCursorPos(void)
{
	return cursorpos;
}

void GUI_TextField::SetCursorPos(int pos)
{
  if(pos >= 0 && pos != cursorpos) {
    if(pos <= strlen(buffer))
  	  this->cursorpos = pos;
    else
  	  this->cursorpos = strlen(buffer);

    // Make sure cursor stays visible
    if(cursorpos == 0)
      startoffset = 0;
    else {
      // Get size of text up to cursor position
      char *cursortext = strdup(buffer);
      cursortext[cursorpos] = '\0';
      SDL_Rect textsize = font->GetTextSize(cursortext);

      if(textsize.w >= area.w - (borderx * 2) - 1 || textsize.w <= startoffset) {
        // Cursor will be outside visible area, change offset so that doesn't happen
        startoffset = textsize.w - (area.w - (borderx * 2)) + 3;
      }
      else
        startoffset = 0;
    }

    MarkChanged();
  }
}

void GUI_TextField::SetBorderX(int pixels)
{
  this->borderx = pixels;
  MarkChanged();
}

void GUI_TextField::SetBorderY(int pixels)
{
  this->bordery = pixels;
  MarkChanged();
}

void GUI_TextField::SetNormalImage(GUI_Surface *surface)
{
	if (GUI_ObjectKeep((GUI_Object **) &background_image, surface))
		MarkChanged();
}

void GUI_TextField::SetFocusImage(GUI_Surface *surface)
{
	if (GUI_ObjectKeep((GUI_Object **) &background_focus_image, surface))
		MarkChanged();
}

void GUI_TextField::SetBackgroundStyle(int style)
{
  if(style == IMAGESTYLE_NORMAL || style == IMAGESTYLE_TILED) {
    background_style = style;
    MarkChanged();
  }
}

void GUI_TextField::SetChangedCallback(GUI_Callback *callback)
{
	GUI_ObjectKeep((GUI_Object **) &changed_callback, callback);
}


void GUI_TextField::SetValidChars(const char *validchars)
{
  free(this->validchars);
	this->validchars = strdup(validchars);
}

const char *GUI_TextField::GetValidChars(void)
{
	return validchars;
}
