/***************************************************************************
                 ifsdl_gui.cpp  -  SDL_gui interface plugin for eboxy
                             -------------------
    begin                : Sun Jul 13 2003
    copyright            : (C) 2003 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <iostream>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "SDL.h"
#include "SDL_thread.h"

#include "ifsdl_gui.h"
#include "../../errors.h"
#include "../../basewidgets/utils.h"
#include "../../constants.h"            // FIXME

#include "buttonwidget.h"
#include "textfieldwidget.h"
#include "labelwidget.h"
#include "togglebuttonwidget.h"
#include "picturewidget.h"
#include "scrollbarwidget.h"
#include "listboxwidget.h"
#include "timer.h"

// the class factories

extern "C" InterfacePlugin *create() {
  return new SDL_guiInterface;
}

extern "C" void destroy(InterfacePlugin *p) {
  delete p;
}
   
using namespace std;

SDL_guiInterface::SDL_guiInterface(void) {
  AddAttribute("width",  anytovariant(640));
  AddAttribute("height", anytovariant(480));
  AddAttribute("depth",  anytovariant(16));
  AddAttribute("fullscreen",  anytovariant(false));
  AddAttribute("hidecursor",  anytovariant(false));
  AddAttribute("nokeyrepeat",  anytovariant(false));
#ifdef JOYSTICK
  AddAttribute("nojoystick",  anytovariant(false));
#endif
  currentPage = NULL;
  loadedPage = NULL;
}

SDL_guiInterface::~SDL_guiInterface(void) {

}

int SDL_guiInterface::Initialise(void) {
	// Initialize SDL
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0) {
    cerr << "Failed to initialise SDL: " << SDL_GetError() << endl;
		return 1;
  }

#ifdef JOYSTICK
  if(!GetAttributeValue("nojoystick")) {
    if(SDL_InitSubSystem(SDL_INIT_JOYSTICK) < 0) {
      cerr << "Failed to initialise SDL joystick subsystem: " << SDL_GetError() << endl;
    }
    joystick = SDL_JoystickOpen(0);
    if(!joystick)
      cerr << "Failed to initialise SDL joystick: " << SDL_GetError() << endl;
    SDL_JoystickEventState(SDL_ENABLE);
  }
#endif

	// Initialize the TTF library
	if (TTF_Init() < 0) {
    cerr << "Failed to initialise TTF font library" << endl;
    SDL_Quit();
		return 1;
  }

	// Initialize the GUI library
	if (GUI_Init() < 0) {
    cerr << "Failed to initialise SDL_gui" << endl;
    TTF_Quit();
    SDL_Quit();
		return 1;
  }

  // Create and install the screen (this will attempt to change to the
  // specified video mode)
  int w = GetAttributeValue("width");
  int h = GetAttributeValue("height");
  int d = GetAttributeValue("depth");
  hidecursor = GetAttributeValue("hidecursor");
  nokeyrepeat = GetAttributeValue("nokeyrepeat");

  opts = SDL_HWSURFACE;
  bool fullscreen = GetAttributeValue("fullscreen");
  if(fullscreen)
    opts |= SDL_FULLSCREEN;

	screen = GUI_ScreenCreate(w, h, d, opts);
	if(screen == NULL) {
    cerr << "Couldn't initialise SDL video mode " << w << "x" <<  h << "x" << d << endl;
    GUI_Quit();
    TTF_Quit();
    SDL_Quit();
		return 1;
  }
 	GUI_SetScreen(screen);

  if(!nokeyrepeat) {
    // Set keyboard repeat
    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
  }

  // Set window caption
  SDL_WM_SetCaption("eboxy", "eboxy");

  // Set up the card stack for easier page changes
  cards = new GUI_CardStack("_cardstack", 0, 0, w, h);
  screen->SetContents(cards);

  if(hidecursor) {
    SDL_ShowCursor(0);
    SDL_WarpMouse(0,0);
  }

  // Tell SDL_gui what thread has access to the gui
  GUI_SetThread(SDL_ThreadID());
  // Start the GUI
  GUI_SetRunning(1);
  
  return 0;
}

void SDL_guiInterface::BeginEventLoop(void (*loopfunc)(void), bool(*interfaceeventfunc)(InterfaceEvent&)) {
  signalquit = false;
  bool nokeypassthrough = GetAttributeValue("nokeypassthrough");
  bool showkeys = GetAttributeValue("showkeys");

  // Process events (mainly interested in keyboard events here)
  while (GUI_GetRunning()) {
    SDL_Event event;
    while (SDL_PollEvent(&event)) {
      bool processed = false;
      if (event.type == SDL_KEYDOWN) {
        if(showkeys) {
          cout << "key scan=" << (int)event.key.keysym.scancode << " sym=" << event.key.keysym.sym;
          cout << " SDLname=\"" << SDL_GetKeyName(event.key.keysym.sym) << "\"" << endl;
        }

        processed = true;
        InterfaceEvent ie;
        ie.type = EIT_KEYDOWN;
        ie.key = transkey(event.key.keysym.sym);
        ie.keymod = transmod(event.key.keysym.mod);
        processed = interfaceeventfunc(ie);

        if(signalquit)
          break;

        // Handle key passthrough (for text-enabled widgets)
        if(!nokeypassthrough && event.key.keysym.unicode != 0)
          SendKeyEvent(event);
      }
#ifdef JOYSTICK
      else if (event.type == SDL_JOYBUTTONDOWN) {
        InterfaceEvent ie;
        ie.type = EIT_JOYBUTTONDOWN;
        if(event.jbutton.button < 8) {
          switch(event.jbutton.button) {
            case 0:
              ie.key = EKEY_JOY_BUTTON_1;
              break;
            case 1:
              ie.key = EKEY_JOY_BUTTON_2;
              break;
            case 2:
              ie.key = EKEY_JOY_BUTTON_3;
              break;
            case 3:
              ie.key = EKEY_JOY_BUTTON_4;
              break;
            case 4:
              ie.key = EKEY_JOY_BUTTON_5;
              break;
            case 5:
              ie.key = EKEY_JOY_BUTTON_6;
              break;
            case 6:
              ie.key = EKEY_JOY_BUTTON_7;
              break;
            case 7:
              ie.key = EKEY_JOY_BUTTON_8;
              break;
            case 8:
              ie.key = EKEY_JOY_BUTTON_9;
              break;
            default:
              ie.key = EKEY_BACKSPACE;
          }
          if(ie.key != EKEY_BACKSPACE)
            processed = interfaceeventfunc(ie);
        }
      }
      else if (event.type == SDL_JOYAXISMOTION) {
        InterfaceEvent ie;
        ie.type = EIT_JOYAXISMOTION;
        ie.key = EKEY_BACKSPACE;
        if(event.jaxis.axis == 0) {
          if(event.jaxis.value > 10)
            ie.key = EKEY_JOY_RIGHT;
          else if(event.jaxis.value < -10)
            ie.key = EKEY_JOY_LEFT;
        }
        else if(event.jaxis.axis == 1) {
          if(event.jaxis.value > 10)
            ie.key = EKEY_JOY_DOWN;
          else if(event.jaxis.value < -10)
            ie.key = EKEY_JOY_UP;
        }
        if(ie.key != EKEY_BACKSPACE)
          processed = interfaceeventfunc(ie);
      }
      else if (event.type == SDL_JOYHATMOTION) {
        InterfaceEvent ie;
        ie.type = EIT_JOYHATMOTION;
        ie.key = EKEY_BACKSPACE;
        switch(event.jhat.value) {
          case SDL_HAT_UP:
            ie.key = EKEY_JOY_HAT_UP;
            break;
          case SDL_HAT_DOWN:
            ie.key = EKEY_JOY_HAT_DOWN;
            break;
          case SDL_HAT_LEFT:
            ie.key = EKEY_JOY_HAT_LEFT;
            break;
          case SDL_HAT_RIGHT:
            ie.key = EKEY_JOY_HAT_RIGHT;
            break;
        }
        if(ie.key != EKEY_BACKSPACE)
          processed = interfaceeventfunc(ie);
      }
#endif

      if(signalquit)
        break;

      if(!processed) {
        // Event not handled here, pass it on
        GUI_Lock();
        screen->Event(&event, 0, 0);
        GUI_Unlock();
      }

      if(!hidecursor && currentPage != NULL && event.type == SDL_MOUSEMOTION)
        currentPage->UpdateFocus();

    }

    loopfunc();

    if(signalquit)
      break;

    GUI_Lock();
    screen->DoUpdate(0);
    GUI_Unlock();

    // Don't hog the CPU
    SDL_Delay(20);
  }
}

void SDL_guiInterface::Quit(void) {
  signalquit = true;
}

void SDL_guiInterface::Uninitialise(void) {
  if(hidecursor)
    SDL_ShowCursor(1);

  ClearPages(true);
    
  screen->SetContents(NULL);
  cards->DecRef();
  screen->DecRef();

  delete currentPage;
  currentPage = NULL;

  resourcemanager.UnloadAllResources();

#ifdef JOYSTICK
  if(joystick)
    SDL_JoystickClose(joystick);
#endif
  GUI_Quit();
  TTF_Quit();
  SDL_Quit();
}

void SDL_guiInterface::ClearPages(bool clearall) {
  ebPage *page;

  for(int i=pages.size()-1;i>=0;i--) {
    page = pages[i];
    if(clearall || page->GetOrigin() == ORIGIN_FILE) {
      pages.erase(pages.begin() + i);
      if(page != currentPage) {
        page->Uninstantiate();
        delete page;
      }
    }
  }
}

Object *SDL_guiInterface::CreateObject(const char *type, const char *name) {
  if(!strcmp(type, TYPESTR_BUTTON))
    return new ButtonWidget(name, 0, 0, -1, -1);
  else if(!strcmp(type, TYPESTR_LABEL))
    return new LabelWidget(name, 0, 0, 0, 0, "", NULL);
  else if(!strcmp(type, TYPESTR_PICTURE))
    return new PictureWidget(name, 0, 0, NULL);
  else if(!strcmp(type, TYPESTR_LISTBOX))
    return new ListBoxWidget(name, 0, 0, -1, -1, NULL, DEFAULT_FONT_COLOR);
  else if(!strcmp(type, TYPESTR_TOGGLEBUTTON))
    return new ToggleButtonWidget(name, 0, 0, -1, -1);
  else if(!strcmp(type, TYPESTR_SCROLLBAR))
    return new ScrollBarWidget(name, 0, 0, -1, -1);
  else if(!strcmp(type, TYPESTR_TEXTFIELD))
    return new TextFieldWidget(name, 0, 0, -1, -1, NULL, DEFAULT_FONT_COLOR, 255);
  else if(!strcmp(type, TYPESTR_PAGE)) {
    ebPage *page = new ebPage(name, 0, 0, 10, 10);
    pages.push_back(page);
    return page;
  }
  else if(!strcmp(type, TYPESTR_TIMER)) {
    TimerObject *timer = new TimerObject(name, 0, false);
    return timer;
  }
  else {
    cerr << "CreateWidget: Unknown widget type '" << type << "'" << endl;
    return NULL;
  }
}

void SDL_guiInterface::DeleteObject(Object *object) {
  // Check if the object is a page
  for(unsigned int i=0; i<pages.size(); i++) {
    if(object == pages[i]) {
      pages.erase(pages.begin() + i);
      break;
    }
  }
 
  delete object;
}

int SDL_guiInterface::SwitchPage(PageBase *page) {
  ebPage *newpage;

  if(page)
    newpage = dynamic_cast<ebPage *>(page);
  else {
    // Find first file page
    newpage = NULL;
    for(unsigned int i=0;i<pages.size();i++) {
      if(pages[i]->GetOrigin() == ORIGIN_FILE) {
        newpage = pages[i];
        break;
      }
    }
    if(!newpage) {
      // The impossible happened
      cerr << "Unable to find first page to switch to" << endl;
      return EBERR_OBJECT_NOT_FOUND;
    }
  }

  newpage->SetWidth(screen->GetWidth());
  newpage->SetHeight(screen->GetHeight());
  loadedPage = newpage;

  if(currentPage) {
    cards->Next();
    ebPage *oldpage = currentPage;
    cards->RemoveWidget(oldpage->GetInstance());
    oldpage->Uninstantiate();
  }

  GUI_Widget *panel = newpage->Instantiate();
  cards->AddWidget(panel);

  currentPage = newpage;

  return EBERR_SUCCESS;
}

PageBase *SDL_guiInterface::GetCurrentPage(void) {
  return currentPage;
}

PageBase *SDL_guiInterface::GetLoadedPage(void) {
  return loadedPage;
}

ResourceManagerBase *SDL_guiInterface::GetResourceManager(void) {
  return &resourcemanager;
}

PageBase *SDL_guiInterface::FindPage(const char *name) {
  for(unsigned int i=0;i<pages.size(); i++) {
    if(!strcasecmp(pages[i]->GetName(), name)) {
      return pages[i];
    }
  }
  return NULL;
}

DynamicObject *SDL_guiInterface::FindExtPageObject(const char *name) {
  for(unsigned int i=0;i<pages.size(); i++) {
    DynamicObject *obj = pages[i]->FindObject(name, NULL);
    if(obj)
      return obj;
  }
  return NULL;
}

int SDL_guiInterface::GetPageCount(void) {
  return pages.size();
}

PageBase *SDL_guiInterface::GetPage(unsigned int index) {
  if(index >= 0 && index < pages.size()) {
    return pages[index];
  }
  else
    return NULL;
}

bool SDL_guiInterface::HasFocus(void) {
  return (SDL_GetAppState() & SDL_APPINPUTFOCUS) != 0;
}

void SDL_guiInterface::SendKeyEvent(SDL_Event e) {
  WidgetBase *widget = currentPage->GetFocusedWidget();

  if(widget != NULL) {
    if(strcmp(widget->GetType(), TYPESTR_TEXTFIELD) == 0) {
      TextFieldWidget *textfield = dynamic_cast<TextFieldWidget *>(widget);
      textfield->SendEvent(e);
    }
  }
}

int SDL_guiInterface::GetScreenWidth(void) {
  return screen->GetWidth();
}

int SDL_guiInterface::GetScreenHeight(void) {
  return screen->GetHeight();
}


// FIXME: move these somewhere appropriate
int transmod(SDLMod sdlmod) {
  int mappedmod;

  mappedmod = 0;
  for(int i=0; modmap[i].modcode != MODKEY_NONE; i++) {
    if((sdlmod & modmap[i].sdlmod) != 0)
      mappedmod |= modmap[i].modcode;
  }

  return mappedmod;
}

eboxykey transkey(SDLKey sdlkey) {
  for(int i=0; keymap[i].key != EKEY_NONE; i++) {
    if(keymap[i].sdlkey == sdlkey)
      return keymap[i].key;
  }
  return EKEY_NONE;
}

