/***************************************************************************
                          pluginmanager.cpp - eboxy plugin manager
                             -------------------
    begin                : Thu Sep 5 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstring>
#include <iostream>
#include <dlfcn.h>
#include <pthread.h>
#include <semaphore.h>
#include <limits.h>

#include "basewidgets/widgetbase.h"
#include "basewidgets/utils.h"
#include "pluginmanager.h"
#include "pluginconstants.h"
#include "plugintypes.h"
#include "errors.h"
#include "basewidgets/timerbase.h"
#include "systemobject.h"
#include "skinparser.h"
#include "main.h"
#include "scriptobject.h"
#include "basewidgets/plugincustomobject.h"
#include "interfacemanager.h"

using namespace std;

// External references to variables defined in main.cpp
extern vector<PageBase *> dynpages;
extern vector<WidgetBase *> tempwidgets;
extern SystemObject *systemObject;
extern PluginManager plugin_manager;
extern InterfaceManager interface_manager;

pthread_mutex_t plugin_mutex;

PluginManager::PluginManager(void) {
  pthread_mutex_init(&plugin_mutex, NULL);
  requestedunload = false;
}

PluginManager::~PluginManager(void) {
  pthread_mutex_destroy(&plugin_mutex);
}

int PluginManager::LoadPlugin(const char *filename, bool initialise) {
  return LoadPlugin(filename, initialise, NULL);
}

int PluginManager::LoadPlugin(const char *filename, bool initialise, Plugin **pluginref) {
  static int plugin_id = 0;
  char libfile[PATH_MAX];
  void *libhandle = NULL;

  // Build path to library
  snprintf(libfile, PATH_MAX, "%s/generic/%s%s", PLUGIN_LIB_DIR, filename, PLUGIN_LIB_SUFFIX);

  // Make sure library exists
  if(!checkFileExists(libfile))
    return EBERR_FILE_ERROR;

  // Load the library
  libhandle = dlopen(libfile, RTLD_LAZY);
  if(libhandle == 0) {
    // Load failed
    char *errorstr = dlerror();
    cerr << "Couldn't load plugin " << libfile << endl;
    cerr << errorstr << endl;
    return EBERR_DLOPEN_FAILED;  // Failed to load library
  }

  // Check the version of the plugin's statically-linked client library
  eboxy_version_info (*fp_clientlib_getver)(void) = (eboxy_version_info (*)(void))dlsym(libhandle, "eboxy_clientlib_getver");
  eboxy_version_info clientlibver = fp_clientlib_getver();
  int rc = checkVersion(clientlibver, true);
  if(rc==1) {
    cerr << "Plugin " << filename << " was compiled for a newer version of eboxy." << endl;
    cerr << "Please upgrade eboxy - you will need at least version " << clientlibver.major << "." << clientlibver.minor << "." << clientlibver.subminor << " to use this plugin." << endl;
    dlclose(libhandle);
    return EBERR_VERSION_MISMATCH;
  }
  else if(rc==-1) {
    cerr << "Plugin " << filename << " was compiled for an older (incompatible) version of eboxy." << endl;
    cerr << "Please obtain an updated version of this plugin." << endl;
    dlclose(libhandle);
    return EBERR_VERSION_MISMATCH;
  }
  
  // Lock mutex
  pthread_mutex_lock(&plugin_mutex);

  // Give the plugin an ID and save it in the plugins list
  plugin_id++;
  Plugin *plugin = new Plugin(filename, plugin_id, libhandle);
  plugins.push_back(plugin);
  if(pluginref)
    *pluginref = plugin;

  // Unlock mutex
  pthread_mutex_unlock(&plugin_mutex);

  if(initialise) {
    return InitialisePlugin(plugin);
  }
  else {
    // Success!
    return EBERR_SUCCESS;
  }
}

int PluginManager::InitialisePlugin(Plugin *plugin) {
  int rc=0;

  plugin->BeforeInitialise();

  // Initialise the plugin
  rc = plugin->Initialise();
  if(rc) {
    // Init failed
    cerr << "Plugin load failed" << endl;
    // Lock mutex
    pthread_mutex_lock(&plugin_mutex);
    // Remove plugin
    for(unsigned int i=0;i<plugins.size();i++) {
      if(plugins[i] == plugin) {
        plugins.erase(plugins.begin() + i);
        break;
      }
    }
    // Unlock mutex
    pthread_mutex_unlock(&plugin_mutex);
    // Unregister objects owned by plugin
    UnregisterPluginObjects(plugin);
    // Dispose of plugin object and return error code
    delete plugin;
    return rc;
  }

  plugin->AfterInitialise();

  // Start the plugin
  plugin->Message(PLMSG_PLUGINSTART, NULL, NULL);

  return EBERR_SUCCESS;
}

void PluginManager::UninitialisePlugin(Plugin *plugin) {
  // Get plugin wrapper to uninitialise
  plugin->Uninitialise();
  // Unregister objects owned by this plugin
  UnregisterPluginObjects(plugin);
}

void PluginManager::UnregisterPluginObjects(Plugin *plugin) {
  // Unregister objects owned by this plugin
  unsigned int i=0;
  while(i<objects.size()) {
    if(objects[i]->GetOwner() == plugin) {
      // Remove object
      delete objects[i];
      objects.erase(objects.begin() + i);
    }
    else
      i++;
  }
}

void PluginManager::UnloadAllPlugins(bool unloadsystem) {
  pthread_mutex_lock(&plugin_mutex);
  for(int i=plugins.size()-1;i>=0;i--) {
    Plugin *plugin = plugins[i];
    if(unloadsystem || plugin->GetOrigin() == PLUGIN_ORIGIN_FILE) {
      // Deinitialise and destruct plugin
      UninitialisePlugin(plugin);
      delete plugin;
      // Remove the pointer from the vector
      plugins.erase(plugins.begin() + i);
    }
  }
  pthread_mutex_unlock(&plugin_mutex);
}

void PluginManager::UnloadAllCustomObjects(bool unloadsystem) {
  // FIXME use unloadsystem properly
  if(unloadsystem) {
    while(objects.size() > 0) {
      CustomObject *object = objects.back();
      delete object;
      objects.pop_back();
    }
  }
}

CustomObject *PluginManager::FindCustomObject(const char *name) {
  for(unsigned int i=0; i<objects.size(); i++) {
    if(!strcasecmp(objects[i]->GetName(), name))
      return objects[i];
  }
  return NULL;
}

WidgetBase *PluginManager::FindTempWidget(const char *name) {
  for(unsigned int i=0; i<tempwidgets.size(); i++) {
    if(!strcasecmp(tempwidgets[i]->GetName(), name))
      return tempwidgets[i];
  }
  return NULL;
}

// FIXME this should be replaced with vectorFind once that actually works
int PluginManager::FindDynamicPage(PageBase *page) {
  for(unsigned int i=0; i<dynpages.size(); i++) {
    if(dynpages[i] == page)
      return i;
  }
  return -1;
}

DynamicObject *PluginManager::FindObject(const char *name) {
  DynamicObject *obj;

  PageBase *loadedpage = getLoadedPage();

  // Check for aliases
  if(!strcasecmp(name, "currentpage")) {
    return loadedpage;
  }
  if(!strcasecmp(name, "focusedwidget")) {
    if(loadedpage)
      return loadedpage->GetFocusedWidget();
    else
      return NULL;
  }
  if(!strcasecmp(name, "system"))
    return systemObject;

  // First look for a custom object
  obj = plugin_manager.FindCustomObject(name);
  if(!obj) {
    // Failed, look for a widget/owned object (only widgets on the current page are accessible)
    obj = findPageObject(name, NULL);
    if(!obj) {
      // Failed, look for a system-level object
      obj = systemObject->FindObject(name, NULL);
      if(!obj) {
        // Failed, look for a dynamically-created widget
        obj = FindTempWidget(name);
        if(!obj) {
          // Failed, look for a page (includes dynamically-created pages)
          obj = interface_manager.FindPage(name);
          if(!obj) {
            // Failed, look for a widget on another page
            obj = interface_manager.FindExtPageObject(name);
            if(!obj) {
              // I think we can safely say the named object doesn't exist at this point :)
              // We don't display an error because this is an internal procedure
              return NULL;
            }
          }
        }
      }
    }
  }
  return obj;
}

Plugin *PluginManager::FindPlugin(int plugin_id) {
  // Lock mutex
  pthread_mutex_lock(&plugin_mutex);

  for(unsigned int i=0; i<plugins.size(); i++) {
    if(plugins[i]->GetID() == plugin_id) {
      // Unlock mutex
      pthread_mutex_unlock(&plugin_mutex);
      return plugins[i];
    }
  }

  // Unlock mutex
  pthread_mutex_unlock(&plugin_mutex);
  return NULL;
}

Plugin *PluginManager::FindPlugin(const char *name) {
  // Lock mutex
  pthread_mutex_lock(&plugin_mutex);

  for(unsigned int i=0; i<plugins.size(); i++) {
    if(!strcmp(plugins[i]->GetName(), name)) {
      // Unlock mutex
      pthread_mutex_unlock(&plugin_mutex);
      return plugins[i];
    }
  }

  // Unlock mutex
  pthread_mutex_unlock(&plugin_mutex);
  return NULL;
}

void PluginManager::BroadcastMessage(int msgcode, void *msgdata1, void *msgdata2) {
  // FIXME: This is a hack.
  // Lock mutex
  pthread_mutex_lock(&plugin_mutex);
  // Unlock mutex
  pthread_mutex_unlock(&plugin_mutex);

  // Send a message to all loaded plugins
  for(unsigned int i=0; i<plugins.size(); i++)
    plugins[i]->Message(msgcode, msgdata1, msgdata2);

  // Check to see if any plugins requested to be unloaded
  UnloadRequestingPlugins();
}

void PluginManager::UnloadRequestingPlugins(void) {
  // Lock mutex
  pthread_mutex_lock(&plugin_mutex);

  if(requestedunload) {
    // At least one plugin has asked to be unloaded already, so find and
    // unload all plugins that have asked to be unloaded
    unsigned int i=0;
    while(i<plugins.size()) {
      if(plugins[i]->IsRequestingUnload()) {
        UninitialisePlugin(plugins[i]);
        delete plugins[i];
        plugins.erase(plugins.begin() + i);
      }
      else
        i++;
    }
  }

  requestedunload = false;

  // Unlock mutex
  pthread_mutex_unlock(&plugin_mutex);
}

int PluginManager::RegisterObject(const char *name, Plugin *plugin) {
  if(FindObject(name) != NULL) {
    cerr << "RegisterObject: Object named " << name << " already exists" << endl;
    return EBERR_OBJECT_EXISTS;
  }
  if(!checkName(name)) {
    cerr << "RegisterObject: \'" << name << "\' is not a valid object name" << endl;
    return EBERR_INVALID_NAME;
  }
  CustomObject *obj = new PluginCustomObject(name);
  obj->SetOwner(plugin);
  objects.push_back(obj);
  plugin->ObjectRegistered(obj);
  return EBERR_SUCCESS;
}

int PluginManager::UnregisterObject(const char *name, Plugin *plugin) {
  unsigned int i=0;

  while(i<objects.size()) {
    if(!strcasecmp(objects[i]->GetName(), name)) {
      if(objects[i]->GetOwner() == plugin) {
        delete objects[i];
        objects.erase(objects.begin() + i);
        return EBERR_SUCCESS;
      }
      else {
        cerr << "fireEvent: Plugin << " << plugin->GetName() << " requested unregistration of object " << name << " which doesn't belong to it" << endl;
        return EBERR_NOT_OWNED;
      }
    }
    else
      i++;
  }

  cerr << "UnregisterObject: Object named " << name << " doesn't exist" << endl;
  return EBERR_OBJECT_NOT_FOUND;
}

void PluginManager::RequestUnload(int plugin_id) {
  // Plugin is asking to be unloaded
  // We have to do this asynchronously - we can't just "pull the plug" on the plugin
  // while one of its functions is still executing.
  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "RequestUnload: Invalid plugin ID " << plugin_id << endl;
    return;
  }

  pthread_mutex_lock(&plugin_mutex);
  requestedunload = true;
  plugin->RequestUnload();
  pthread_mutex_unlock(&plugin_mutex);
}

void PluginManager::PrintPluginList(void) {
  // Print a list of loaded plugins

  // Lock mutex
  pthread_mutex_lock(&plugin_mutex);

  // Print list of plugins
  cout << plugins.size() << " loaded plugin(s) - active:" << endl;
  for(unsigned int i=0; i<plugins.size(); i++) {
    if(plugins[i]->GetState() == PLSTATE_RUNNING) {
      const plugin_info *pinfo = plugins[i]->GetInfo();
      cout << plugins[i]->GetName() << PLUGIN_LIB_SUFFIX << ": " << pinfo->namestring << " " << pinfo->verstring << "\t[" << plugins[i]->GetState() << "]" << endl;
    }
  }

  // Unlock mutex
  pthread_mutex_unlock(&plugin_mutex);
}

Variant PluginManager::CallMethod(const char *object, const char *method, unsigned int numargs, Variant args[]) {
  DynamicObject *obj = plugin_manager.FindObject(object);
  if(obj)
    return obj->MethodCall(method, numargs, args);
  else {
    cerr << "CallMethod: No object named " << object << " exists" << endl;
    return VARNULL;
  }
}

Variant PluginManager::GetPropertyValue(const char *object, const char *property) {
  DynamicObject *obj = plugin_manager.FindObject(object);
  if(obj)
    return obj->GetPropertyValue(property);
  else {
    cerr << "GetPropertyValue: No object named " << object << " exists" << endl;
    return VARNULL;
  }
}

int PluginManager::SetPropertyValue(const char *object, const char *property, Variant value) {
  DynamicObject *obj = plugin_manager.FindObject(object);
  if(obj)
    return obj->SetPropertyValue(property, value);
  else {
    cerr << "SetPropertyValue: No object named " << object << " exists" << endl;
    return -1;
  }
}

extern "C" {
  // The following functions are exposed to the plugin client library (and hence to the plugins)

void *resolveFunction(const char *name) {
  // This function is used by the plugin client library to get pointers to all of the API functions
  if(strcmp(name, "findObject") == 0)
    return (void *)*findObject;
  else if(strcmp(name, "typeOfObject") == 0)
    return (void *)*typeOfObject;
  else if(strcmp(name, "getPropertyAsString") == 0)
    return (void *)*getPropertyAsString;
  else if(strcmp(name, "setPropertyAsString") == 0)
    return (void *)*setPropertyAsString;
  else if(strcmp(name, "callMethod") == 0)
    return (void *)*callMethod;
  else if(strcmp(name, "registerObject") == 0)
    return (void *)*registerObject;
  else if(strcmp(name, "registerPropertyDL") == 0)
    return (void *)*registerPropertyDL;
  else if(strcmp(name, "registerEventHandlerDL") == 0)
    return (void *)*registerEventHandlerDL;
  else if(strcmp(name, "registerMethodDL") == 0)
    return (void *)*registerMethodDL;
  else if(strcmp(name, "fireEvent") == 0)
    return (void *)*fireEvent;
  else if(strcmp(name, "getPluginInfo") == 0)
    return (void *)*getPluginInfo;
  else if(strcmp(name, "setPluginInfo") == 0)
    return (void *)*setPluginInfo;
  else if(strcmp(name, "requestUnload") == 0)
    return (void *)*requestUnload;
  else if(strcmp(name, "changePage") == 0)
    return (void *)*changePage;
  else if(strcmp(name, "loadXMLFile") == 0)
    return (void *)*loadXMLFile;
  else if(strcmp(name, "performAction") == 0)
    return (void *)*performActionAPI;
  else if(strcmp(name, "unregisterObject") == 0)
    return (void *)*unregisterObject;
  else if(strcmp(name, "unregisterEventHandlerDL") == 0)
    return (void *)*unregisterEventHandlerDL;
  else if(strcmp(name, "createWidget") == 0)
    return (void *)*createWidget;
  else if(strcmp(name, "createPage") == 0)
    return (void *)*createPage;
  else if(strcmp(name, "cloneWidget") == 0)
    return (void *)*cloneWidget;
  else if(strcmp(name, "deleteWidget") == 0)
    return (void *)*deleteWidget;
  else if(strcmp(name, "deletePage") == 0)
    return (void *)*deletePage;
  else if(strcmp(name, "runScript") == 0)
    return (void *)*runScript;
  else
    return NULL;
}

int findObject(const char *object, const char *type) {
  if(!object || strlen(object) == 0)
    return 0;

  DynamicObject *obj = plugin_manager.FindObject(object);
  if(obj) {
    if(type == NULL || !(strcmp(obj->GetType(), type)))
      return 1;
  }

  return 0;
}

const char *typeOfObject(const char *object) {
  DynamicObject *obj = plugin_manager.FindObject(object);
  if(obj)
    return obj->GetType();
  else {
    cerr << "typeOfObject: No object named " << object << " exists" << endl;
    return NULL;
  }
}

char *getPropertyAsString(const char *object, const char *property) {
  Variant value = plugin_manager.GetPropertyValue(object, property);
  if(value==VARNULL)
    return NULL;
  else
    return strdup(value.c_str());
}

int setPropertyAsString(const char *object, const char *property, const char *value) {
  DynamicObject *obj = plugin_manager.FindObject(object);
  if(obj) {
    return obj->SetPropertyValue(property, anytovariant(value));
  }
  else {
    cerr << "setPropertyAsString: No object named " << object << " exists" << endl;
    return -1;
  }
}

char *callMethod(const char *object, const char *method, int numargs, const char *args[]) {
  if(numargs < 0) {
    cerr << "callMethod: numargs must be positive" << endl;
    return NULL;
  }

  Variant vargs[numargs];
  for(int j=0;j<numargs;j++)
    vargs[j] = args[j];

  Variant retval = plugin_manager.CallMethod(object, method, numargs, vargs);

  if(retval==VARNULL)
    return NULL;
  else
    return strdup(retval.c_str());
}

int registerObject(const char *name, int plugin_id) {
  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "registerObject: Invalid plugin ID " << plugin_id << endl;
    return EBERR_INVALID_PLUGIN_ID;
  }

  return plugin_manager.RegisterObject(name, plugin);
}

int registerPropertyDL(const char *object, const char *property, int plugin_id, const char *getter, const char *setter) {
  propertygetfn getterfunction = NULL;
  propertysetfn setterfunction = NULL;

  CustomObject *obj = plugin_manager.FindCustomObject(object);
  if(!obj) {
    cerr << "registerPropertyDL: No object named " << object << " exists" << endl;
    return EBERR_OBJECT_NOT_FOUND;
  }

  if(!checkName(property)) {
    cerr << "registerPropertyDL: \'" << property << "\' is not a valid property name" << endl;
    return EBERR_INVALID_NAME;
  }

  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "registerPropertyDL: Invalid plugin ID " << plugin_id << endl;
    return EBERR_INVALID_PLUGIN_ID;
  }

  if(obj->GetOwner() != plugin) {
    cerr << "registerPropertyDL: Plugin << " << plugin->GetName() << " requested addition of a property to " << object << " which doesn't belong to it" << endl;
    return EBERR_NOT_OWNED;
  }

  if(getter)
    getterfunction = (propertygetfn)plugin->GetFunction(getter);
  if(setter)
    setterfunction = (propertysetfn)plugin->GetFunction(setter);

  return obj->AddProperty(property, getterfunction, setterfunction);
}

int registerEventHandlerDL(const char *object, const char *event, int plugin_id, const char *eventfuncname) {
  int (*eventfunction)(const char *) = NULL;

  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "registerEventHandlerDL: Invalid plugin ID " << plugin_id << endl;
    return EBERR_INVALID_PLUGIN_ID;
  }

  if(!checkName(event)) {
    cerr << "registerEventHandlerDL: \'" << event << "\' is not a valid event name" << endl;
    return EBERR_INVALID_NAME;
  }

  eventfunction = (int (*)(const char *))plugin->GetFunction(eventfuncname);
  if(!eventfunction) {
    cerr << "registerEventHandlerDL: failed to get specified function " << eventfuncname << endl;
    return EBERR_MISSING_FUNCTION;
  }

  DynamicObject *obj = plugin_manager.FindObject(object);
  if(obj) {
    EventHandler *handler = obj->RegisterEventHandler(event, plugin, eventfunction);
    plugin->AddOwnedEventHandler(handler);
  }
  else {
    cerr << "registerEventHandlerDL: No object named " << object << " exists" << endl;
    return EBERR_OBJECT_NOT_FOUND;
  }
  return EBERR_SUCCESS;
}

int registerMethodDL(const char *object, const char *method, int plugin_id, int numargs, const char *methodfuncname) {
  methodfn methodfunction = NULL;

  if(numargs < 0) {
    cerr << "registerMethodDL: numargs must be positive" << endl;
    return EBERR_INVALID_ARGS;
  }

  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "registerMethodDL: Invalid plugin ID " << plugin_id << endl;
    return EBERR_INVALID_PLUGIN_ID;
  }

  if(!checkName(method)) {
    cerr << "registerMethodDL: \'" << method << "\' is not a valid method name" << endl;
    return EBERR_INVALID_NAME;
  }

  // Use dlsym() to get a pointer to the function inside the plugin
  methodfunction = (methodfn)plugin->GetFunction(methodfuncname);
  if(!methodfunction) {
    cerr << "registerMethodDL: failed to get specified function " << methodfuncname << endl;
    return EBERR_MISSING_FUNCTION;
  }

  // Look for the specified custom object
  CustomObject *obj = plugin_manager.FindCustomObject(object);
  if(obj) {
    if(obj->GetOwner() != plugin) {
      cerr << "registerMethodDL: Plugin << " << plugin->GetName() << " requested addition of a method to " << object << " which doesn't belong to it" << endl;
      return EBERR_NOT_OWNED;
    }
    // Check if there's a property by that name already (for script sanity purposes)
    if(obj->HasProperty(method)) {
      cerr << "registerMethodDL: Object " << object << " already has a property named " << method << endl;
      return EBERR_MEMBER_EXISTS;
    }
    if(obj->HasMethod(method)) {
      cerr << "registerMethodDL: Object " << object << " already has a method named " << method << endl;
      return EBERR_MEMBER_EXISTS;
    }
    MethodHandler *handler = obj->RegisterMethod(method, plugin, numargs, methodfunction);
    plugin->AddOwnedMethodHandler(handler);
  }
  else {
    cerr << "registerMethodDL: No custom object named " << object << " exists" << endl;
    return EBERR_OBJECT_NOT_FOUND;
  }

  return EBERR_SUCCESS;
}

int fireEvent(const char *object, const char *event, int plugin_id) {
  CustomObject *obj = plugin_manager.FindCustomObject(object);
  if(!obj) {
    cerr << "fireEvent: No object named " << object << " exists" << endl;
    return EBERR_OBJECT_NOT_FOUND;
  }

  if(!checkName(event)) {
    cerr << "fireEvent: \'" << event << "\' is not a valid event name" << endl;
    return EBERR_INVALID_NAME;
  }

  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "fireEvent: Invalid plugin ID " << plugin_id << endl;
    return EBERR_INVALID_PLUGIN_ID;
  }

  if(obj->GetOwner() != plugin) {
    cerr << "fireEvent: Plugin << " << plugin->GetName() << " requested firing of an event on " << object << " which doesn't belong to it" << endl;
    return EBERR_NOT_OWNED;
  }

  // Fire the event
  return obj->FireEvent(event);
}

const plugin_info *getPluginInfo(const char *name) {
  Plugin *plugin = plugin_manager.FindPlugin(name);
  if(plugin)
    return plugin->GetInfo();
  else {
    cerr << "getPluginInfo: plugin " << name << " not currently loaded" << endl;
    return NULL;
  }
}

void setPluginInfo(int plugin_id, const char *namestring, const char *verstring) {
  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "registerMethodDL: Invalid plugin ID " << plugin_id << endl;
    return;
  }

  plugin->SetInfo(namestring, verstring);
}

void requestUnload(int plugin_id) {
  // Plugin is asking to be unloaded
  plugin_manager.RequestUnload(plugin_id);
}

int changePage(const char *pagename) {
  return switchPageEvent(pagename);
}

int loadXMLFile(const char *filename, const char *pagename) {
  return switchFileEvent(filename, pagename);
}

int performActionAPI(const char *actionname) {
  return performAction(actionname);
}

int unregisterObject(const char *name, int plugin_id) {
  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "unregisterObject: Invalid plugin ID " << plugin_id << endl;
    return EBERR_INVALID_PLUGIN_ID;
  }

  return plugin_manager.UnregisterObject(name, plugin);
}

int unregisterEventHandlerDL(const char *object, const char *event, int plugin_id) {
  Plugin *plugin = plugin_manager.FindPlugin(plugin_id);
  if(!plugin) {
    cerr << "unregisterEventHandlerDL: Invalid plugin ID " << plugin_id << endl;
    return EBERR_INVALID_PLUGIN_ID;
  }

  DynamicObject *obj = plugin_manager.FindObject(object);
  if(obj) {
    int rc = obj->UnregisterEventHandler(event, plugin);
    if(rc == EBERR_EVENT_NOT_HANDLED) {
      cerr << "unregisterEventHandlerDL: no handler for event " << event << " was registered" << endl;
      return EBERR_EVENT_NOT_HANDLED;
    }
  }
  else {
    cerr << "unregisterEventHandlerDL: No object named " << object << " exists" << endl;
    return EBERR_OBJECT_NOT_FOUND;
  }
  return EBERR_SUCCESS;
}

int createWidget(const char *name, const char *type, const char *templatename) {
  // Check to see if widget already exists
  if(plugin_manager.FindObject(name) != NULL) {
    cerr << "createWidget: an object with the name " << name << " already exists" << endl;
    return EBERR_OBJECT_EXISTS;
  }

  // Create the new widget
  WidgetBase *widget = CreateWidget(name, type);
  if(widget) {
    if(templatename) {
      int ret = applyTemplate(widget, templatename);
      if(ret != 0) {
        cerr << "createWidget: failed to apply template " << templatename << " to widget" << endl;
        delete widget;
        return ret;
      }
    }
    // Add widget to temp list
    tempwidgets.push_back(widget);
  }
  else {
    cerr << "createWidget: failed to create object of type " << type << endl;
    return EBERR_CREATE_FAILED;
  }

  return EBERR_SUCCESS;
}

int createPage(const char *name) {
  if(plugin_manager.FindObject(name) != NULL) {
    cerr << "createPage: an object with the name " << name << " already exists" << endl;
    return EBERR_OBJECT_EXISTS;
  }

  // Get interface plugin to create page
  InterfacePlugin *ifplugin = interface_manager.GetPlugin();
  PageBase *page = dynamic_cast<PageBase *>(ifplugin->CreateObject("page", name));
  if(page) {
    dynpages.push_back(page);
  }
  else {
    cerr << "createPage: failed to create page object" << endl;
    return EBERR_CREATE_FAILED;
  }

  return EBERR_SUCCESS;
}

int cloneWidget(const char *original, const char *duplicate) {
  // Find the original
  DynamicObject *origobj = plugin_manager.FindObject(original);
  if(origobj == NULL) {
    cerr << "cloneWidget: no object with the name " << original << " found" << endl;
    return EBERR_OBJECT_NOT_FOUND;
  }
  // Make sure the destination doesn't exist
  if(plugin_manager.FindObject(duplicate) != NULL) {
    cerr << "cloneWidget: Object named " << duplicate << " already exists" << endl;
    return EBERR_OBJECT_EXISTS;
  }

  // Create the new widget
  WidgetBase *widget = CreateWidget(duplicate, origobj->GetType());
  if(widget) {
    // Copy data to new widget
    
    WidgetBase *origwidget = dynamic_cast<WidgetBase *>(origobj);
    try {
      origwidget->Clone(widget);
    }
    catch(int e) {
      // Failed to clone properties
      cerr << "cloneWidget: failed to clone object (error " << e << ")" << endl;
      delete widget;
      return e;
    }
    // Add widget to temp list
    tempwidgets.push_back(widget);
  }
  else {
    cerr << "cloneWidget: failed to create object of type " << origobj->GetType() << endl;
    return EBERR_CREATE_FAILED;
  }

  return EBERR_SUCCESS;
}

int deleteWidget(const char *name) {
  for(unsigned int i=0;i<tempwidgets.size();i++) {
    if(!strcasecmp(tempwidgets[i]->GetName(), name)) {
      // Check to see if widget is instantiated
      if(tempwidgets[i]->IsInstantiated()) {
        cerr << "deleteWidget: Widget named " << name << " is in use" << endl;
        return EBERR_OBJECT_IN_USE;
      }
      // Remove from temp widgets vector
      WidgetBase *widget = tempwidgets[i];
      vector<WidgetBase *>::iterator p = tempwidgets.begin() + i;
      tempwidgets.erase(p);
      // Get interface plugin to delete widget
      InterfacePlugin *ifplugin = interface_manager.GetPlugin();
      ifplugin->DeleteObject(widget);
      // Done
      return EBERR_SUCCESS;
    }
  }

  // Didn't find it
  cerr << "deleteWidget: Unable to find widget named " << name << " in constructed widgets" << endl;
  return EBERR_OBJECT_NOT_FOUND;
}

int deletePage(const char *name) {
  InterfacePlugin *ifplugin = interface_manager.GetPlugin();
  PageBase *page = ifplugin->FindPage(name);
  if(page) {
    // Check to see if page is instantiated
    if(page->IsInstantiated()) {
      cerr << "deletePage: page named " << name << " is in use" << endl;
      return EBERR_OBJECT_IN_USE;
    }
    // Remove from dynamic pages vector
    int index = plugin_manager.FindDynamicPage(page);
    if(index == -1) {
      cerr << "deletePage: cannot delete page " << name << " - not dynamically created" << endl;
      return EBERR_OBJECT_NOT_FOUND;
    }
    vector<PageBase *>::iterator p = dynpages.begin() + index;
    dynpages.erase(p);
    // Get interface plugin to delete page
    ifplugin->DeleteObject(page);
    return EBERR_SUCCESS;
  }
  else {
    // Didn't find it
    cerr << "deletePage: Unable to find page named " << name << endl;
    return EBERR_OBJECT_NOT_FOUND;
  }
}

int runScript(const char *code, int flags) {
  Script script("runtimescript", code, flags); // FIXME: name
  script.Run();
  // FIXME: Script return code?
  return EBERR_SUCCESS;
}

}
