/***************************************************************************
                  pluginobject.cpp - eboxy-side plugin wrapper object
                             -------------------
    begin                : Sat Oct 5 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <dlfcn.h>
#include <pthread.h>

#include "pluginobject.h"
#include "pluginmanager.h"
#include "pluginconstants.h"
#include "errors.h"
#include "main.h"

using namespace std;

extern const char *TYPESTR_PLUGIN;
extern PluginManager plugin_manager;

Plugin::Plugin(const char *name, int id, void *library)
: Object(name) {
  state = PLSTATE_CONSTRUCTING;
  this->library = library;
  this->plugin_id = id;
  this->requestedunload = false;
  this->type = TYPESTR_PLUGIN;
  this->findable = false; // we don't want any calls to FindObject to find this object
  this->info.namestring = NULL;
  this->info.verstring = NULL;
  pthread_mutex_init(&init_mutex, NULL);
  state = PLSTATE_DORMANT;
  origin = PLUGIN_ORIGIN_SYSTEM;
}

Plugin::~Plugin(void) {
  if(!(state == PLSTATE_NONEXISTENT || state == PLSTATE_DORMANT)) {
    Uninitialise();
  }

  // Close the library
  if(dlclose(library))
    cerr << "Failed to unload shared library " << name << endl;

  // Unregister all event handlers registered on objects owned by this plugin
  while(myeventhandlers.size() > 0) {
    EventTemplate *evt = myeventhandlers.back();
    // Remove from the list
    myeventhandlers.pop_back();
    free(evt->type);
    evt->script->DecRef();
    // We don't get rid of the script - that is deleted elsewhere
    delete evt;
  }

  pthread_mutex_destroy(&init_mutex);
}

void Plugin::BeforeInitialise(void) {
  // This happens before Initialise is called
  pthread_mutex_lock(&init_mutex);
}

void Plugin::AfterInitialise(void) {
  // This happens after Initialise is called
  pthread_mutex_unlock(&init_mutex);
}

int Plugin::Initialise(void) {
  int rc = 0;
  int (*fp_ebplugin_init)(void) = NULL;
  int (*fp_ebplugin_message)(int, void *, void *) = NULL;
  void (*fp_ebplugin_deinit)(void) = NULL;
  int (*fp_clientlib_init)(void *(*)(const char *), int) = NULL;
  void (*fp_clientlib_deinit)(void) = NULL;
  eboxy_version_info (*fp_clientlib_getver)(void) = NULL;

  // Get pointers to the standard plugin functions
  // FIXME: should probably show error messages returned by dlerror() here
  fp_ebplugin_init      = (int (*)(void))dlsym(library, "ebplugin_init");
  fp_ebplugin_message   = (int (*)(int, void *, void *))dlsym(library, "ebplugin_message");
  fp_ebplugin_deinit    = (void (*)(void))dlsym(library, "ebplugin_deinit");
  fp_clientlib_init     = (int (*)(void *(*)(const char *), int))dlsym(library, "eboxy_clientlib_init");
  fp_clientlib_deinit   = (void (*)(void))dlsym(library, "eboxy_clientlib_deinit");
  fp_clientlib_getver   = (eboxy_version_info (*)(void))dlsym(library, "eboxy_clientlib_getver");

  // Check that we got pointers to all of the required functions
  if(fp_ebplugin_init == 0 || fp_ebplugin_message == 0 || fp_ebplugin_deinit == 0) {
    // Missing one of the user implemented functions
    cerr << "Plugin " << name << " does not seem to implement required functions - contact plugin author" << endl;
    return EBERR_MISSING_FUNCTION;
  }
  if(fp_clientlib_init == 0 || fp_clientlib_deinit == 0 || fp_clientlib_getver == 0) {
    cerr << "Plugin " << name << " does not seem to be linked properly - contact plugin author" << endl;
    return EBERR_MISSING_CLIENTLIB;
  }

  // Call the plugin client library init function
  rc = fp_clientlib_init(*resolveFunction, plugin_id);
  if(rc) {
    // Client library init failed
    // FIXME: need to decide if it's really safe to continue here or not
    cerr << "Failed to initialise plugin " << name << " (client lib returned " << rc << ")" << endl;
    return rc;
  }

  // Call the plugin's init function
  rc = fp_ebplugin_init();
  if(rc) {
    // Plugin init failed
    // FIXME: need to decide if it's really safe to continue here or not
    cerr << "Failed to initialise plugin " << name << " (plugin returned " << rc << ")" << endl;
    return EBERR_PLUGIN_INIT_FAILED;
  }

  // Save function pointers
  this->ebplugin_message = fp_ebplugin_message;
  this->ebplugin_deinit = fp_ebplugin_deinit;
  this->clientlib_deinit = fp_clientlib_deinit;

  state = PLSTATE_RUNNING;

  // Success
  return EBERR_SUCCESS;
}

int Plugin::Uninitialise(void) {
  if(state > PLSTATE_INITIALISING) {
    state = PLSTATE_DEINITIALISING;
    // Call the plugin deinitialisation function
    ebplugin_deinit();
    // Call the client library deinitialisation function
    clientlib_deinit();
  }
  state = PLSTATE_DESTRUCTING;

  // FIXME: In case you can't tell, this is a horrible hack (tries to make sure mutex is unlocked
  // and waiting code executes. I am unsure as to whether it really works or not).
  pthread_mutex_trylock(&init_mutex);
  pthread_mutex_unlock(&init_mutex);

  // Unregister all event handlers registered by this plugin
  while(ownedeventhandlers.size() > 0) {
    EventHandler *handler = ownedeventhandlers.back();
    // Remove from the list
    ownedeventhandlers.pop_back();
    // Unregister the event handler, which deletes it
    DynamicObject *dobj = dynamic_cast<DynamicObject *>(handler->GetSource());
    dobj->UnregisterEventHandler(handler);
  }

  // Unregister all methods registered by this plugin
  while(ownedmethodhandlers.size() > 0) {
    MethodHandler *handler = ownedmethodhandlers.back();
    // Remove from the list
    ownedmethodhandlers.pop_back();
    // Unregister the method, which deletes it
    DynamicObject *dobj = dynamic_cast<DynamicObject *>(handler->GetOwner());
    dobj->UnregisterMethod(handler);
  }

  if(info.namestring != NULL) {
    free(info.namestring);
    info.namestring = NULL;
  }
  if(info.verstring != NULL) {
    free(info.verstring);
    info.verstring = NULL;
  }

  state = PLSTATE_DORMANT;

  return EBERR_SUCCESS;
}

int Plugin::Message(int msgcode, void *msgdata1, void *msgdata2) {
  // Wait until the plugin has finished initialising
  pthread_mutex_lock(&init_mutex);
  pthread_mutex_unlock(&init_mutex);
  if(state == PLSTATE_DESTRUCTING || state == PLSTATE_DEINITIALISING) {
    // Time to get out of here, quick, the ceiling is caving in!
    return 0;
  }

  if(state == PLSTATE_DORMANT)
    return 0;

  // Send the message to the plugin
  return ebplugin_message(msgcode, msgdata1, msgdata2);
}

int Plugin::GetState(void) {
  return state;
}

int Plugin::GetID(void) {
  return plugin_id;
}

const plugin_info *Plugin::GetInfo(void) {
  return &info;
}

void Plugin::SetInfo(const char *namestring, const char *verstring) {
  if(info.namestring) {
    cerr << "Plugin error: setPluginInfo may only be called once" << endl;
    return;
  }
  info.namestring = strdup(namestring);
  info.verstring = strdup(verstring);
}

void *Plugin::GetFunction(const char *funcname) {
  return dlsym(library, funcname);
}

void Plugin::RequestUnload(void) {
  requestedunload = true;
}

bool Plugin::IsRequestingUnload(void) {
  return requestedunload;
}

void Plugin::AddEventHandler(EventTemplate *evt) {
  evt->script->IncRef();
  myeventhandlers.push_back(evt);
}

void Plugin::RemoveEventHandler(EventTemplate *evt) {
  unsigned int i=0;
  while(i<myeventhandlers.size()) {
    if(myeventhandlers[i] == evt) {
      myeventhandlers.erase(myeventhandlers.begin() + i);
      free(evt->type);
      evt->script->DecRef();
      delete evt;
      break;
    }
    else
      i++;
  }
}

void Plugin::ObjectRegistered(CustomObject *obj) {
  // Called by plugin manager when the plugin registers an object
  for(unsigned int i=0; i<myeventhandlers.size(); i++) {
    if(!strcasecmp(obj->GetName(), myeventhandlers[i]->sourceobject)) {
      EventHandler *handler = obj->RegisterEventHandler(myeventhandlers[i]->type, myeventhandlers[i]->script);
      handler->SetThreaded(myeventhandlers[i]->threaded);
    }
  }
}

void Plugin::AddOwnedEventHandler(EventHandler *handler) {
  ownedeventhandlers.push_back(handler);
}

void Plugin::AddOwnedMethodHandler(MethodHandler *handler) {
  ownedmethodhandlers.push_back(handler);
}

void Plugin::RemoveOwnedEventHandler(EventHandler *handler) {
  unsigned int i=0;
  while(i<ownedeventhandlers.size()) {
    if(ownedeventhandlers[i] == handler) {
      ownedeventhandlers.erase(ownedeventhandlers.begin() + i);
      break;
    }
    else
      i++;
  }
}

void Plugin::RemoveOwnedMethodHandler(MethodHandler *handler) {
  unsigned int i=0;
  while(i<ownedmethodhandlers.size()) {
    if(ownedmethodhandlers[i] == handler) {
      ownedmethodhandlers.erase(ownedmethodhandlers.begin() + i);
      break;
    }
    else
      i++;
  }
}

int Plugin::DoInitialise(void) {
  // Initialise the plugin
  return plugin_manager.InitialisePlugin(this);
}

void Plugin::DoUninitialise(void) {
  // Uninitialise the plugin
  return plugin_manager.UninitialisePlugin(this);
}

void Plugin::RemoveOwnedObject(void *object) {
  // Called by an object that a plugin has registered a handler on
  for(unsigned int i=0; i<ownedeventhandlers.size(); i++) {
    if(ownedeventhandlers[i] == object) {
      RemoveOwnedEventHandler(ownedeventhandlers[i]);
      return;
    }
  }
  for(unsigned int i=0; i<ownedmethodhandlers.size(); i++) {
    if(ownedmethodhandlers[i] == object) {
      RemoveOwnedMethodHandler(ownedmethodhandlers[i]);
      return;
    }
  }
}

void Plugin::SetOrigin(int origin) {
  this->origin = origin;
}

int Plugin::GetOrigin(void) {
  return origin;
}
