/***************************************************************************
               ebscript_flex.l  -  flex input for eboxy script parser
                             -------------------                                         
    begin                : Sat Sep 7 2002                                           
    copyright            : (C) 2002 by Paul Eggleton                         
    email                : bluelightning@bluelightning.org                                     
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   * 
 *                                                                         *
 ***************************************************************************/

%option noyywrap
%option yylineno
%option c++
%option yyclass="ScriptLexer"

%{
#include <iostream>
#include <sstream>  // Deprecated, but required for flex string parsing
#include <cstring>

#include "script.h"

#define YY_FATAL_ERROR(msg) printScriptError((ScriptLexer *)this, msg);

char string_buf[MAX_STR_CONST];
char *string_buf_ptr;
YY_BUFFER_STATE last_buf_state;

%}

ID      [a-z_][a-z0-9_]*
DIGIT   [0-9]

%x str

%%
\"      {
          string_buf_ptr = string_buf;
          BEGIN(str);
        }

<str>\" { /* closing quote */
          BEGIN(INITIAL);
          *string_buf_ptr = '\0';
          semantic_value.string = strdup(string_buf); /* Don't forget to free! */
          return tSTRING;
        }
<str>\n {
          std::cerr << "Script error: Unterminated string constant at line " << yylineno << std::endl;
          BEGIN(INITIAL);
        }
<str>\\n  *string_buf_ptr++ = '\n';
<str>\\t  *string_buf_ptr++ = '\t';
<str>\\r  *string_buf_ptr++ = '\r';
<str>\\b  *string_buf_ptr++ = '\b';
<str>\\f  *string_buf_ptr++ = '\f';
<str>\\(.|\n)  *string_buf_ptr++ = yytext[1];
<str>[^\\\n\"]+ {
          char *yptr = yytext;
          while ( *yptr )
            *string_buf_ptr++ = *yptr++;
        }

\-?{DIGIT}+ {
              semantic_value.number = atoi(yytext);
              return tNUMBER;
            }
true        return tTRUE;
false       return tFALSE;
and         return tAND;
or          return tOR;
xor         return tXOR;
not         return tNOT;
null        return tNULL;
echo        return tECHO;
exechide    return tEXECHIDE;
execwait    return tEXECWAIT;
execwaitcapture return tEXECWAITCAPTURE;
exec        return tEXEC;
load        return tLOAD;
loadplugin  return tLOADPLUGIN;
page        return tPAGE;
quit        return tQUIT;
delay       return tDELAY;
==          return tEQUAL;
\!=         return tNOTEQUAL;
\<          return tLESS;
\>          return tGREATER;
\<=         return tLESSEQUAL;
\>=         return tGREATEREQUAL;
=           return tASSIGN;
\+          return tPLUS;
\-          return tMINUS;
\*          return tMULTIPLY;
\/          return tDIVIDE;
%           return tREMAINDER;
\^          return tPOWER;
\.          return tDOT;
\(          return tLBRACKET;
\)          return tRBRACKET;
\,          return tCOMMA;
{ID}        {
              semantic_value.string = strdup(yytext);  /* Don't forget to free! */
              return tIDENT;
            }
#[^\n]*     return tCOMMENT; /* this allows us to have completely commented out scripts */
[ \t\n]+    /* eat whitespace */
.           {
              std::string errmsg("unexpected character ");
              errmsg += yytext;
              printScriptError(this, errmsg);
            }
%%

void ScriptLexer::embedded_yylex(int& result, YYSTYPE *lvalp){
  result = yylex();
  *lvalp = semantic_value;
}

ScriptLexer::ScriptLexer(char *string, int flags) {
  // Initialise variables
  this->lineoffset = 0;
  this->filename = NULL;
  this->flags = flags;
  // Load string into stream and parse stream
  mem_input = new std::istringstream(string);
  switch_streams(mem_input, NULL);
}

ScriptLexer::~ScriptLexer(void) {
  if(this->filename)
    free(this->filename);
  delete mem_input;
}

void ScriptLexer::SetLineOffset(int line) {
  this->lineoffset = line;
}

int ScriptLexer::GetLineNumber(void) {
  return lineno() + lineoffset - 1;
}

void ScriptLexer::SetFileName(const char *filename) {
  if(this->filename)
    free(this->filename);
  this->filename = strdup(filename);
}

const char *ScriptLexer::GetFileName(void) {
  return filename;
}

bool ScriptLexer::IsTrusted(void) {
  return flags & SCRIPTFLAGS_TRUSTED == SCRIPTFLAGS_TRUSTED;
}

void printScriptError(ScriptLexer *lexer, std::string message) {
  // Print out line number and file
  int errline = lexer->lineno() - 1;
  int xmlerrline = lexer->GetLineNumber();
  const char *xmlfilename = lexer->GetFileName();
  std::cerr << "Script error at line " << errline;
  if(xmlfilename)
    std::cerr << " (" << xmlfilename << ", line " << xmlerrline << ")" << std::endl;
  else
    std::cerr << std::endl;
  if(message != "")
    std::cerr << "  " << message << endl;
}
