/***************************************************************************
                          skinparser.cpp  -  XML skin file parser
                             -------------------
    begin                : Thu Nov 7 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdio>
#include <iostream>
#include <cstring>
#include <cstdlib>
#include <vector>

#include "libxml/xmlmemory.h"
#include "libxml/parser.h"
#include "libxml/tree.h"
#include "unistd.h"

#include "basewidgets/widgetbase.h"
#include "basewidgets/utils.h"
#include "basewidgets/variant.h"
#include "errors.h"

#include "basewidgets/pagebase.h"
#include "basewidgets/timerbase.h"
#include "systemobject.h"
#include "pluginmanager.h"
#include "pluginconstants.h"
#include "scriptobject.h"
#include "basewidgets/scripteventhandler.h"
#include "main.h"
#include "interfacemanager.h"
#include "basewidgets/resourcemanagerbase.h"
#include "basewidgets/resource.h"
#include "basewidgets/actionmanager.h"

#include "skinparser.h"

using namespace std;

extern SystemObject *systemObject;
extern PluginManager plugin_manager;
extern const char *DTD_FILE;
extern InterfaceManager interface_manager;
extern ResourceManagerBase *resourcemanager;
extern vector<ActionHandler *> globalactionhandlers;
extern ActionManager *actionmanager;
extern const char *TYPESTR_TIMER;

vector<GenericTemplate *> templates;

int parseSkin(std::string &filename, bool system) {
  xmlDocPtr doc;
  xmlNodePtr cur;
  char *str;
  int rc = 0;

  if(!checkFileExists(filename.c_str())) // FIXME
    return EBERR_FILE_ERROR;

  // Build a DOM tree from the XML file
  xmlLineNumbersDefault(1);
  doc = xmlParseFile(filename.c_str());
  if (doc == NULL) {
    cerr << "Failed to parse XML file" << endl;
    return EBERR_XML_PARSE_FAILED;
  }

  // Change to the directory where the XML file is (for image/font paths)
  string filepath = getFilePath(filename);
  chdir(filepath.c_str());

  // Check the document "type"
  cur = xmlDocGetRootElement(doc);
  if (cur == NULL) {
    cerr << "XML file " << filename << " is empty" << endl;
    xmlFreeDoc(doc);
  	return EBERR_XML_EMPTY;
  }
  if (xmlStrcmp(cur->name, (const xmlChar *) "eboxy")) {
    cerr << "XML file " << filename << " has the wrong type (root node should be \'eboxy\')" << endl;
  	xmlFreeDoc(doc);
  	return EBERR_XML_NOT_VALID;
  }

  str = (char *)xmlGetProp(cur, (const xmlChar *) "minversion");
  if(str) {
    // Minimum version specified - verify it
    string majorver, minorver, subminorver;
    int count = 0;

    if(!checkRegExp(str, "[0-9]+\\.[0-9]+\\.[0-9]+")) {
      cerr << "XML file " << filename << ": attribute 'minversion' is not of the correct format (n.n.n)" << endl;
      free(str);
      xmlFreeDoc(doc);
      return EBERR_XML_NOT_VALID;
    }

    for(int i=0;i<strlen(str);i++) {
      if(str[i] == '.')
        count++;
      else {
        switch(count) {
          case 0:
            majorver += str[i];
            break;
          case 1:
            minorver += str[i];
            break;
          case 2:
            subminorver += str[i];
            break;
        }
      }
    }

    eboxy_version_info verinfo;

    verinfo.major = strtoint(majorver);
    verinfo.minor = strtoint(minorver);
    verinfo.subminor = strtoint(subminorver);

    if(checkVersion(verinfo, false)) {
      cerr << "XML file " << filename << " is intended for use with a newer version of eboxy" << endl;
      cerr << "You will need at least version " << str << endl;
      free(str);
      xmlFreeDoc(doc);
      return EBERR_VERSION_MISMATCH;
    }
        
    free(str);
  }

  // Verify the XML file against the DTD
  xmlDtdPtr dtd = xmlParseDTD(NULL, (const xmlChar *) DTD_FILE);
  if (dtd == NULL) {
    xmlGenericError(xmlGenericErrorContext, "Could not parse DTD %s\n", DTD_FILE);
    xmlFreeDoc(doc);
    xmlCleanupParser();
    return EBERR_DTD_PARSE_FAILED;
  } else {
    xmlValidCtxt cvp;
    cvp.userData = (void *) stderr;
    // FIXME: this doesn't print out the XML line number the error occurred in
    cvp.error = (xmlValidityErrorFunc) fprintf;
    cvp.warning = (xmlValidityWarningFunc) fprintf;
    if (!xmlValidateDtd(&cvp, doc, dtd)) {
      xmlGenericError(xmlGenericErrorContext, "XML file %s does not validate against DTD %s\n", filename.c_str(), DTD_FILE);
      xmlFreeDtd(dtd);
      xmlFreeDoc(doc);
      xmlCleanupParser();
      return EBERR_XML_NOT_VALID;
    }
  }
  xmlFreeDtd(dtd);

  // Walk the tree
  cur = cur->xmlChildrenNode;
  while(cur) {
    if(!xmlIsBlankNode(cur)) {
      if (!(xmlStrcmp(cur->name, (const xmlChar *) "templates")) ) {
        rc = parseTemplates(doc, cur, system);
        if(rc != EBERR_SUCCESS) {
          // Error occurred, message has already been displayed
          xmlFreeDoc(doc);
          xmlCleanupParser();
          return rc;
        }
      }
      if (!(xmlStrcmp(cur->name, (const xmlChar *) "pages")) ) {
        rc = parsePages(doc, cur, system);
        if(rc != EBERR_SUCCESS) {
          // Error occurred, message has already been displayed
          xmlFreeDoc(doc);
          xmlCleanupParser();
          return rc;
        }
      }
      if (!(xmlStrcmp(cur->name, (const xmlChar *) "system")) ) {
        rc = parseSystem(doc, cur, system);
        if(rc != EBERR_SUCCESS) {
          // Error occurred, message has already been displayed
          xmlFreeDoc(doc);
          xmlCleanupParser();
          return rc;
        }
      }
    }
    cur = cur->next;
  }

  // Done parsing
  xmlFreeDoc(doc);
  xmlCleanupParser();

  // See what we've got
  if(!system && interface_manager.GetPageCount() == 0) {
    cerr << "XML file " << filename << " does not contain any pages" << endl;
    return EBERR_XML_NOPAGES;
  }

  return EBERR_SUCCESS;
}

int parsePages(xmlDocPtr doc, xmlNodePtr node, bool system) {
  xmlNodePtr cur;

  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if (!xmlStrcmp(cur->name, (const xmlChar *) "page")) {
      PageBase *page = parsePage(doc, cur, system);
      if(page) {
        if(system)
          page->SetOrigin(ORIGIN_SYSTEM);
        else
          page->SetOrigin(ORIGIN_FILE);
      }
      else
        return EBERR_XML_BAD;
    }
    cur = cur->next;
  }
  return EBERR_SUCCESS;
}

int parseTemplates(xmlDocPtr doc, xmlNodePtr node, bool system) {
  xmlNodePtr cur;

  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    // New parsing method
    if(cur->type == XML_ELEMENT_NODE) {
      GenericTemplate *gtemplate = parseTemplate(doc, cur);
      if(gtemplate) {
        if(system)
          gtemplate->origin = ORIGIN_SYSTEM;
        else
          gtemplate->origin = ORIGIN_FILE;
        templates.push_back(gtemplate);
      }
      else
        return EBERR_XML_BAD;
    }
    cur = cur->next;
  }
  return EBERR_SUCCESS;
}

int parseSystem(xmlDocPtr doc, xmlNodePtr node, bool system) {
  xmlNodePtr cur;

  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    // FIXME not checking for errors here  
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "event")) ) {
      parseEvent(doc, cur, systemObject);
    }
    else if ((!xmlStrcmp(cur->name, (const xmlChar *) "timer")) ) {
      TimerObjectBase *timer = parseTimer(doc, cur);
      if(timer)
        systemObject->AddObject(timer);
    }
    else if ((!xmlStrcmp(cur->name, (const xmlChar *) "plugin")) ) {
      Plugin *plugin = parsePlugin(doc, cur, system);
      if(plugin)
        systemObject->AddUnownedObject(plugin); // plugin is managed automatically by the plugin manager
    }
    else if ((!xmlStrcmp(cur->name, (const xmlChar *) "actionhandler")) ) {
      ActionHandler *handler = parseActionHandler(doc, cur, system);
      if(handler) {
        // FIXME: move this somewhere more appropriate
        int rc = actionmanager->RegisterGlobalAction(handler->GetAction(), handler);
        if(rc != EBERR_SUCCESS) {
          // Couldn't register action
          cout << "debug error 1" << endl;
          delete handler;
          return rc;
        }
        globalactionhandlers.push_back(handler);
      }
    }
    cur = cur->next;
  }
  return EBERR_SUCCESS;
}

PageBase *parsePage(xmlDocPtr doc, xmlNodePtr node, bool system) {
  xmlNodePtr cur;
  char *str;
  PageBase *page;
  ImageResource *background;
  char *name;

  str = (char *)xmlGetProp(node, (const xmlChar *) "name");
  if(str) {
    if(!checkName(str)) {
      cerr << "\'" << str << "\' is not a valid name (line " << xmlGetLineNo(node) << ")" << endl;
      free(str);
      return NULL;
    }
    name = strdup(str);
    free(str);
  }

  str = (char *)xmlGetProp(node, (const xmlChar *) "background");
  if(str) {
    background = resourcemanager->LoadImageResource(str, false);
    free(str);
    if(background == NULL) {
      // Error occurred, and message has already been displayed
      return NULL;
    }
  }
  else
    background = NULL;

  InterfacePlugin *plugin = interface_manager.GetPlugin();
  page = dynamic_cast<PageBase *>(plugin->CreateObject("page", name));


  page->SetBackground(background);
  free(name); // important

  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if(cur->type == XML_ELEMENT_NODE) {
      if ((!xmlStrcmp(cur->name, (const xmlChar *) "plugin")) ) {
        Plugin *plugin = parsePlugin(doc, cur, system);
        if(plugin)
          page->AddUnownedObject(plugin); // plugin is managed automatically by the plugin manager
      }
      else if ((!xmlStrcmp(cur->name, (const xmlChar *) "event")) ) {
        parseEvent(doc, cur, page);
      }
      else if ((!xmlStrcmp(cur->name, (const xmlChar *) "timer")) ) {
        TimerObjectBase *timer = parseTimer(doc, cur);
        if(timer)
          page->AddObject(timer);
        else
          return NULL;
      }
      else if ((!xmlStrcmp(cur->name, (const xmlChar *) "actionhandler")) ) {
        ActionHandler *handler = parseActionHandler(doc, cur, system);
        if(handler)
          page->AddActionHandler(handler);
        else
          return NULL;
      }
      else {
        // A widget
        WidgetBase *widget = parseWidget(doc, cur);
        if(widget) {
          if(system)
            widget->SetOrigin(ORIGIN_SYSTEM);
          else
            widget->SetOrigin(ORIGIN_FILE);
          page->AddWidget(widget);
        }
        else
          return NULL;
      }
    }
    cur = cur->next;
  }

  return page;
}

WidgetBase *parseWidget(xmlDocPtr doc, xmlNodePtr node) {
  xmlNodePtr cur;
  char *str;
  char *name;
  WidgetBase *widget;
  GenericTemplate *gtemplate = NULL;
  static int count = 0;
  
  // Grab name if specified, otherwise make one up (useful for debugging)
  str = (char *)xmlGetProp(node, (const xmlChar *) "name");
  if(str) {
    if(!checkName(str)) {
      cerr << "\'" << str << "\' is not a valid name (line " << xmlGetLineNo(node) << ")" << endl;
      free(str);
      return NULL;
    }
    name = strdup(str);
    free(str);
  }
  else {
    name = newstring(strlen((const char *)node->name) + 6);
    sprintf(name, "_%s%d", (const char *)node->name, ++count);
  }

  widget = CreateWidget(name, (const char *)node->name);
  if(!widget) {
    free(name);
    return NULL;
  }
  
  // Grab template
  str = (char *)xmlGetProp(node, (const xmlChar *) "template");
  if(str) {
    gtemplate = findTemplate(str);
    if(!gtemplate) {
      cerr << "Non-existent template " << str << " specified (line " << xmlGetLineNo(node) << ")" << endl;
      delete widget;
      free(name);
      free(str);
      return NULL;
    }
    free(str);
  }

  // Set properties
  int propcount = widget->GetPropertyCount();

  int index;
  CustomProperty *prop;
  bool propset;
  for(int i=0;i<propcount;i++) {
    prop = widget->GetProperty(i);
    if(!(prop->IsReadOnly())) {
      propset = false;
      str = (char *)xmlGetProp(node, (const xmlChar *) prop->GetName());
      if(str) {
        int rc = prop->SetValue(anytovariant(str));
        if(rc != EBERR_SUCCESS) {
          // Unable to set property value
          delete widget;
          free(name);
          free(str);
          return NULL;
        }
        propset = true;
        free(str);
      }
      else {
        // Check for template value
        if(gtemplate) {
          index = gtemplate->attributes.IndexOfName(prop->GetName());
          if(index > -1) {
            int rc = prop->SetValue(anytovariant(gtemplate->attributes.GetValue(index)));
            if(rc != EBERR_SUCCESS) {
              // Unable to set property value
              delete widget;
              free(name);
              return NULL;
            }
            propset = true;
          }
        }
      }

      if(!propset && prop->IsRequired()) {
        // Property required and not set
        cerr << "Required property " << prop->GetName() << " of object " << name << " has not been specified (line " << xmlGetLineNo(node) << ")" << endl;
        delete widget;
        free(name);
        return NULL;
      }
    }
  }

  if(widget->GetDefaultEvent()) {
    // Grab default event
    str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    if(str != NULL) {
      if(!isEmptyString(str)) {
        Script *script = new Script("noname", str); // FIXME: name
        script->SetLineOffset(xmlGetLineNo(node));
        widget->RegisterEventHandler(widget->GetDefaultEvent(), script); // FIXME: check if default event exists
        script->DecRef(); // let go of reference (event handler should have called IncRef() already)
      }
      free(str);
    }
  }
  else if(widget->GetDefaultProperty()) {
    // Grab default property (currently only used by labels)
    str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    if(str != NULL) {
      widget->SetPropertyValue(widget->GetDefaultProperty(), anytovariant(str));
      free(str);
    }
  }

  // Grab attached events
  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if (!xmlStrcmp(cur->name, (const xmlChar *) "event")) {
      parseEvent(doc, cur, widget);
    }
    cur = cur->next;
  }

  free(name);
  return widget;
}

void parseEvent(xmlDocPtr doc, xmlNodePtr node, DynamicObject *parent) {
  EventTemplate *evt = parseEvent(doc, node);

  if(evt->script && evt->type) {
    EventHandler *handler = parent->RegisterEventHandler(evt->type, evt->script);
    handler->SetThreaded(evt->threaded);
    evt->script->DecRef(); // let go of reference (event handler should have called IncRef() already)
  }
  else {
    cerr << "Invalid event specified for " << parent->GetName() << " (line " << xmlGetLineNo(node) << ")" << endl;
    if(evt->script)
      delete evt->script;
  }

  if(evt->type)
    free(evt->type);
  delete evt;
}

EventTemplate *parseEvent(xmlDocPtr doc, xmlNodePtr node) {
  EventTemplate *evt;
  char *str;

  evt = new EventTemplate;

  // Grab event type
  evt->type = (char *)xmlGetProp(node, (const xmlChar *) "type");

  str = (char *)xmlGetProp(node, (const xmlChar *) "wait");
  if(str) {
    evt->threaded = !strtobool(str);
    free(str);
  }
  else
    evt->threaded = true;

  // Grab script
  str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
  if(str != NULL && !isEmptyString(str)) {
    evt->script = new Script("script1", str);
    evt->script->SetLineOffset(xmlGetLineNo(node));
  }
  else
    evt->script = NULL;
  if(str)
    free(str);

  return evt;
}

TimerObjectBase *parseTimer(xmlDocPtr doc, xmlNodePtr node) {
  char *str;
  char *name;
  int interval = 1000;
  bool autostart = false;
  TimerObjectBase *timer;
  static int count = 0;

  // Grab name if specified, otherwise make one up (useful for debugging)
  str = (char *)xmlGetProp(node, (const xmlChar *) "name");
  if(str) {
    if(!checkName(str)) {
      cerr << "\'" << str << "\' is not a valid name (line " << xmlGetLineNo(node) << ")" << endl;
      free(str);
      return NULL;
    }
    name = strdup(str);
    free(str);
  }
  else {
    name = newstring(15);
    sprintf(name, "_timer%d", ++count);
  }

  str = (char *)xmlGetProp(node, (const xmlChar *) "interval");
  if(str) {
    interval = atoi(str);
    free(str);
  }

  str = (char *)xmlGetProp(node, (const xmlChar *) "enabled");
  if(str) {
    autostart = strtobool(str);
    free(str);
  }

  // Create the timer
  InterfacePlugin *plugin = interface_manager.GetPlugin();
  timer = dynamic_cast<TimerObjectBase *>(plugin->CreateObject(TYPESTR_TIMER, name));

  // Grab attached event handlers
  str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
  if(str != NULL) {
    if(strlen(str) > 0) {
      Script *script = new Script("noname", str); // FIXME: name
      script->SetLineOffset(xmlGetLineNo(node));
      timer->RegisterEventHandler("OnTimer", script);
      script->DecRef(); // let go of reference (event handler should have called IncRef() already)
    }
    free(str);
  }

  free(name);
  return timer;
}

Plugin *parsePlugin(xmlDocPtr doc, xmlNodePtr node, bool system) {
  char *str;
  xmlNodePtr cur, cur2;
  Plugin *plugin;
  EventTemplate *evt;
  char *objectname;

  str = (char *)xmlGetProp(node, (const xmlChar *) "module");
  if(str) {
    // Load the plugin
    if(plugin_manager.LoadPlugin(str, false, &plugin) != EBERR_SUCCESS) {
      // Error occurred during load (we assume that an error message has been displayed)
      free(str);
      return NULL;
    }
    free(str);

    if(system)
      plugin->SetOrigin(PLUGIN_ORIGIN_SYSTEM);
    else
      plugin->SetOrigin(PLUGIN_ORIGIN_FILE);
  }
  else {
    // DTD shouldn't allow this, but what the heck...
    cerr << "No plugin module specified (line " << xmlGetLineNo(node) << ")" << endl;
    return NULL;
  }

  // Parse contents of plugin object element (currently, only attached events)
  cur = node->xmlChildrenNode;
  while (cur != NULL) {
    if (!xmlStrcmp(cur->name, (const xmlChar *) "pluginobject")) {
      // Get the name of the object we're looking at
      objectname = (char *)xmlGetProp(cur, (const xmlChar *) "name");
      // Get events that the user wants to bind to this object
      cur2 = cur->xmlChildrenNode;
      while (cur2 != NULL) {
        if (!xmlStrcmp(cur2->name, (const xmlChar *) "event")) {
          // Register an event (well, actually the registering gets done later)
          evt = parseEvent(doc, cur2);
          evt->sourceobject = strdup(objectname);
          plugin->AddEventHandler(evt);
          evt->script->DecRef(); // let go of reference (AddEventHandler() should have called IncRef() already)
        }
        cur2 = cur2->next;
      }
      // Release object name string
      free(objectname);
    }
    cur = cur->next;
  }

  return plugin;
}

ActionHandler *parseActionHandler(xmlDocPtr doc, xmlNodePtr node, bool system) {
  char *str;
  char *actionname;
  ActionHandler *handler;
  int rc;

  handler = NULL;
      
  // Grab action handler
  str = (char *)xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
  if(str != NULL) {
    if(strlen(str) > 0) {
      actionname = (char *)xmlGetProp(node, (const xmlChar *) "action");
      if(actionname) {
        Script *script = new Script("noname", str); // FIXME: name
        handler = new ActionHandler(script, actionname, system);
        script->DecRef();
        free(actionname);
      }
    }
    free(str);
  }

  return handler;
}

void clearTemplates(bool clearall) {
  for(int i=templates.size()-1; i>=0; i--) {
    GenericTemplate *templt = templates[i];
    if(clearall || templt->origin == ORIGIN_FILE) {
      templates.erase(templates.begin() + i);
      if(templt->name)
        free(templt->name);
      delete templt;
    }
  }
}

GenericTemplate *parseTemplate(xmlDocPtr doc, xmlNodePtr node) {
  GenericTemplate *gtemplate;

  // Create template object
  gtemplate = new GenericTemplate;
  
  // Set template type
  if (!xmlStrcmp(node->name, (const xmlChar *) "buttontemplate"))
    gtemplate->type = TT_BUTTON;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "labeltemplate"))
    gtemplate->type = TT_LABEL;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "listboxtemplate"))
    gtemplate->type = TT_LISTBOX;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "scrollbartemplate"))
    gtemplate->type = TT_SCROLLBAR;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "textfieldtemplate"))
    gtemplate->type = TT_TEXTFIELD;
  else if (!xmlStrcmp(node->name, (const xmlChar *) "togglebuttontemplate"))
    gtemplate->type = TT_TOGGLEBUTTON;
  else {
    cerr << "parseTemplate: Unknown template type '" << node->name << "'" << endl;
    delete gtemplate;
    return NULL;
  }
  
  // Get attributes into a StringList
  _xmlAttr *prop = node->properties;
  while(prop) {
    if(strcmp((char *)prop->name, "name") == 0)
      gtemplate->name = strdup((char *)prop->children->content);
    else
      gtemplate->attributes.SetValue((char *)prop->name, (char *)prop->children->content);
    prop = prop->next;
  }

  return gtemplate;
}

GenericTemplate *findTemplate(const char *templatename) {
  // Find the referenced template
  for(unsigned int i=0;i<templates.size();i++) {
    if(!strcasecmp(templates[i]->name, templatename)) {
      return templates[i];
    }
  }
  return NULL;
}

int applyTemplate(WidgetBase *widget, const char *templatename) {
  // Find specified template
  GenericTemplate *gtemplate = findTemplate(templatename);
  if(!gtemplate)
    return EBERR_OBJECT_NOT_FOUND;
  
  // Set widget properties from template
  int propcount = widget->GetPropertyCount();
  for(int i=0;i<propcount;i++) {
    CustomProperty *prop = widget->GetProperty(i);
    int index = gtemplate->attributes.IndexOfName(prop->GetName());
    if(index > -1) {
      int rc = prop->SetValue(anytovariant(gtemplate->attributes.GetValue(index)));
      if(rc != EBERR_SUCCESS)
        return rc;
    }
  }

  return EBERR_SUCCESS;
}

WidgetBase *CreateWidget(const char *name, const char *type) {
  InterfacePlugin *plugin = interface_manager.GetPlugin();
  return dynamic_cast<WidgetBase *>(plugin->CreateObject(type, name));
}

